"""

The import wizard helps the user importing CSV-like data from a file into a
numpy array. The wizard containst three pages:

SelectFilePage:
    - The user selects a file and previews its contents (or, the beginning of it)
    
SetParametersPage:
    - The user selects delimiters, etc. and selects which columns to import
    - A preview of the data in tabualar form is shown, with colors indicating
      how the file is parsed: yellow for header rows, green for the comments
      column and red for values that could not be parsed
      
ResultPage:
    - The wizard shows the generated code that is to be used to import the file
      according to the settings
    - The user chooses to execute the code in the current shell or paste the
      code into the editor


"""

import unicodedata
import os.path as op

import pyzo.codeeditor
from . import QtCore, QtGui, QtWidgets
from pyzo import translate


# All keywords in Python 2 and 3. Obtained using: import keyword; keyword.kwlist
# Merged from Py2 and 3
keywords = [
    "False",
    "None",
    "True",
    "and",
    "as",
    "assert",
    "break",
    "class",
    "continue",
    "def",
    "del",
    "elif",
    "else",
    "except",
    "exec",
    "finally",
    "for",
    "from",
    "global",
    "if",
    "import",
    "in",
    "is",
    "lambda",
    "nonlocal",
    "not",
    "or",
    "pass",
    "print",
    "raise",
    "return",
    "try",
    "while",
    "with",
    "yield",
]


class CodeView(
    pyzo.codeeditor.IndentationGuides,
    pyzo.codeeditor.CodeFolding,
    pyzo.codeeditor.Indentation,
    pyzo.codeeditor.HomeKey,
    pyzo.codeeditor.EndKey,
    pyzo.codeeditor.NumpadPeriodKey,
    pyzo.codeeditor.AutoIndent,
    pyzo.codeeditor.PythonAutoIndent,
    pyzo.codeeditor.SyntaxHighlighting,
    pyzo.codeeditor.CodeEditorBase,
):  # CodeEditorBase must be the last one in the list
    """
    Code viewer, stripped down version of the CodeEditor
    """

    pass


class SelectFilePage(QtWidgets.QWizardPage):
    """
    First page of the wizard, select file and preview contents
    """

    def __init__(self):
        QtWidgets.QWizardPage.__init__(self)

        self.setTitle(translate("importwizard", "Select file"))

        self.txtFilename = QtWidgets.QLineEdit()
        self.btnBrowse = QtWidgets.QPushButton(translate("importwizard", "Browse..."))
        self.preview = QtWidgets.QPlainTextEdit()
        self.preview.setReadOnly(True)

        vlayout = QtWidgets.QVBoxLayout()
        hlayout = QtWidgets.QHBoxLayout()

        hlayout.addWidget(self.txtFilename)
        hlayout.addWidget(self.btnBrowse)
        vlayout.addLayout(hlayout)
        vlayout.addWidget(QtWidgets.QLabel(translate("importwizard", "Preview:")))
        vlayout.addWidget(self.preview)

        self.setLayout(vlayout)

        self.registerField("fname", self.txtFilename)

        self.btnBrowse.clicked.connect(self.onBrowseClicked)
        self.txtFilename.editingFinished.connect(self.updatePreview)
        self._isComplete = False

    def onBrowseClicked(self):
        # Difference between PyQt4 and PySide: PySide returns filename, filter
        # while PyQt4 returns only the filename
        filename = QtWidgets.QFileDialog.getOpenFileName(
            filter="Text files (*.txt *.csv);;All files (*.*)"
        )
        if isinstance(filename, tuple):
            filename = filename[0]

        filename = str(filename).replace("/", op.sep)  # Show native file separator

        self.txtFilename.setText(filename)
        self.updatePreview()

    def updatePreview(self):
        filename = self.txtFilename.text()
        if not filename:
            data = ""
            self._isComplete = False
            self.wizard().setPreviewData(None)
        else:
            try:
                with open(filename, "rb") as file:
                    maxsize = 5000
                    data = file.read(maxsize)
                    more = bool(file.read(1))  # See if there is more data available

                data = data.decode("ascii", "replace")

                self.wizard().setPreviewData(data)

                if more:
                    data += "..."

                self._isComplete = True  # Allow to proceed to the next page
            except Exception as e:
                data = str(e)
                self._isComplete = False
                self.wizard().setPreviewData(None)

        self.preview.setPlainText(data)
        self.completeChanged.emit()

    def isComplete(self):
        return self._isComplete


class SetParametersPage(QtWidgets.QWizardPage):
    def __init__(self):
        QtWidgets.QWizardPage.__init__(self)

        self.setTitle("Select parameters")

        self._columnNames = None

        def genComboBox(choices):
            cbx = QtWidgets.QComboBox()
            for choice in choices:
                cbx.addItem(choice)
            cbx.setEditable(True)
            return cbx

        self.cbxDelimiter = genComboBox(",;")
        self.cbxComments = genComboBox("#%'")
        self.sbSkipHeader = QtWidgets.QSpinBox()

        self.preview = QtWidgets.QTableWidget()
        self.preview.setSelectionModel(
            QtCore.QItemSelectionModel(self.preview.model())
        )  # Work-around for reference tracking bug in PySide
        self.preview.setSelectionBehavior(self.preview.SelectColumns)
        self.preview.setSelectionMode(self.preview.MultiSelection)

        # Layout

        formlayout = QtWidgets.QFormLayout()
        formlayout.addRow("Delimiter", self.cbxDelimiter)
        formlayout.addRow("Comments", self.cbxComments)
        formlayout.addRow("Header rows to skip", self.sbSkipHeader)

        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(formlayout)
        layout.addWidget(
            QtWidgets.QLabel(translate("importwizard", "Select columns to import:"))
        )
        layout.addWidget(self.preview)

        self.setLayout(layout)

        # Wizard fields
        self.registerField("delimiter", self.cbxDelimiter, "currentText")
        self.registerField("comments", self.cbxComments, "currentText")
        self.registerField("skip_header", self.sbSkipHeader)

        # Signals
        self.cbxComments.editTextChanged.connect(self.updatePreview)
        self.cbxDelimiter.editTextChanged.connect(self.updatePreview)
        self.sbSkipHeader.valueChanged.connect(self.updatePreview)
        self.preview.verticalHeader().sectionClicked.connect(self.onRowHeaderClicked)

    def columnNames(self):
        if self._columnNames is None:
            return list(
                ["d" + str(i + 1) for i in range(self.preview.columnCount() - 1)]
            )

        return list(self._columnNames)

    def updateHorizontalHeaderLabels(self):
        self.preview.setHorizontalHeaderLabels(self.columnNames() + ["Comments"])

    def onRowHeaderClicked(self, row):
        names = self.parseColumnNames(row)
        self._columnNames = names
        self.updateHorizontalHeaderLabels()

    def parseColumnNames(self, row):
        """
        Use the data in the given row to create column names. First, try the
        data in the data columns. If these are all empty, use the comments
        column, split by the given delimiter.

        Names are fixed up to be valid Python 2 / Python 3 identifiers
        (chars a-z A-Z _ 0-9 , no Python 2 or 3 keywords, not starting with 0-9)

        returns: list of names, exactly as many as there are data columns
        """
        names = []
        columnCount = self.preview.columnCount() - 1
        for col in range(columnCount):
            cell = self.preview.item(row, col)
            if cell is None:
                names.append("")
            else:
                names.append(cell.text().strip())

        # If no values found, try the comments:
        if not any(names):
            cell = self.preview.item(row, columnCount)
            if cell is not None:
                comment = cell.text()[1:].strip()  # Remove comment char and whitespace
                delimiter = self.cbxDelimiter.currentText()
                names = list(name.strip() for name in comment.split(delimiter))

                # Ensure names is exactly columnCount long
                names += [""] * columnCount
                names = names[:columnCount]

        # Fixup names
        def fixname(name, col):
            # Remove accents
            name = "".join(
                c
                for c in unicodedata.normalize("NFD", name)
                if unicodedata.category(c) != "Mn"
            )
            # Replace invalid chars with _
            name = "".join(
                c
                if (c.lower() >= "a" and c.lower() <= "z") or (c >= "0" and c <= "9")
                else "_"
                for c in name
            )

            if not name:
                return "d" + str(col)

            if name[0] >= "0" and name <= "9":
                name = "d" + name

            if name in keywords:
                name = name + "_"

            return name

        names = list(fixname(name, i + 1) for i, name in enumerate(names))

        return names

    def selectedColumns(self):
        """
        Returns a tuple of the columns that are selected, or None if no columns
        are selected
        """
        selected = []
        for selrange in self.preview.selectionModel().selection():
            selected += range(selrange.left(), selrange.right() + 1)

        selected.sort()

        if not selected:
            return None
        else:
            return tuple(selected)

    def initializePage(self):
        self.updatePreview()

    def updatePreview(self):
        # Get settings from the currently specified values in the wizard
        comments = self.cbxComments.currentText()
        delimiter = self.cbxDelimiter.currentText()
        skipheader = self.sbSkipHeader.value()
        if not comments or not delimiter:
            return

        # Store current selection, will be restored at the end
        selectedColumns = self.selectedColumns()

        # Clear the complete table
        self.preview.clear()
        self.preview.setColumnCount(0)
        self.preview.setRowCount(0)

        # Iterate through the source file line by line
        # Process like genfromtxt, with names = False
        # However, we do keep the header lines and comments; we show them in
        # distinct colors so that the user can see how the data is selected

        source = iter(self.wizard().previewData().splitlines())

        def split_line(line):
            """Chop off comments, strip, and split at delimiter."""
            line, sep, commentstr = line.partition(comments)
            line = line.strip(" \r\n")
            if line:
                return line.split(delimiter), sep + commentstr
            else:
                return [], sep + commentstr

        # Insert comments column
        self.preview.insertColumn(0)

        ncols = 0  # Number of columns, excluding comments column
        headerrows = 0  # Number of header rows, including empty header rows

        inheader = True

        for lineno, line in enumerate(source):
            fields, commentstr = split_line(line)

            # Process header like genfromtxt, with names = False
            if lineno >= skipheader and fields:
                inheader = False

            if inheader:
                headerrows = lineno + 1  # +1 since lineno = 0 is the first line

            self.preview.insertRow(lineno)

            # Add columns to fit all fields
            while len(fields) > ncols:
                self.preview.insertColumn(ncols)
                ncols += 1

            # Add fields to the table
            for col, field in enumerate(fields):
                cell = QtWidgets.QTableWidgetItem(field)
                if not inheader:
                    try:
                        float(field)
                    except ValueError:
                        cell.setBackground(QtGui.QBrush(QtGui.QColor("pink")))

                self.preview.setItem(lineno, col, cell)

            # Add the comment
            cell = QtWidgets.QTableWidgetItem(commentstr)
            cell.setBackground(QtGui.QBrush(QtGui.QColor("lightgreen")))

            self.preview.setItem(lineno, ncols, cell)

        # Colorize the header cells. This is done as the last step, since
        # meanwhile new columns (and thus new cells) may have been added
        for row in range(headerrows):
            for col in range(ncols):
                cell = self.preview.item(row, col)
                if not cell:
                    cell = QtWidgets.QTableWidgetItem("")
                    self.preview.setItem(row, col, cell)

                cell.setBackground(QtGui.QBrush(QtGui.QColor("khaki")))

        # Try to restore selection
        if selectedColumns is not None:
            for column in selectedColumns:
                self.preview.selectColumn(column)

        # Restore column names
        self.updateHorizontalHeaderLabels()


class ResultPage(QtWidgets.QWizardPage):
    """
    The resultpage lets the user select wether to import the data as a single
    2D-array, or as one variable (1D-array) per column

    Then, the code to do the import is generated (Py2 and Py3 compatible). This
    code can be executed in the current shell, or copied into the current editor

    """

    def __init__(self):
        QtWidgets.QWizardPage.__init__(self)
        self.setTitle("Execute import")
        self.setButtonText(
            QtWidgets.QWizard.FinishButton, translate("importwizard", "Close")
        )

        self.rbAsArray = QtWidgets.QRadioButton(
            translate("importwizard", "Import data as single array")
        )
        self.rbPerColumn = QtWidgets.QRadioButton(
            translate("importwizard", "Import data into one variable per column")
        )
        self.rbAsArray.setChecked(True)

        self.chkInvalidRaise = QtWidgets.QCheckBox(
            translate("importwizard", "Raise error upon invalid data")
        )
        self.chkInvalidRaise.setChecked(True)

        self.codeView = CodeView()
        self.codeView.setParser("python")
        self.codeView.setZoom(pyzo.config.view.zoom)
        self.codeView.setFont(pyzo.config.view.fontname)

        self.btnExecute = QtWidgets.QPushButton("Execute in current shell")
        self.btnInsert = QtWidgets.QPushButton("Paste into current file")

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.rbAsArray)
        layout.addWidget(self.rbPerColumn)
        layout.addWidget(self.chkInvalidRaise)

        layout.addWidget(QtWidgets.QLabel("Resulting import code:"))
        layout.addWidget(self.codeView)
        layout.addWidget(self.btnExecute)
        layout.addWidget(self.btnInsert)
        self.setLayout(layout)

        self.registerField("invalid_raise", self.chkInvalidRaise)

        self.btnExecute.clicked.connect(self.onBtnExecuteClicked)
        self.btnInsert.clicked.connect(self.onBtnInsertClicked)
        self.rbAsArray.clicked.connect(self.updateCode)
        self.rbPerColumn.clicked.connect(self.updateCode)
        self.chkInvalidRaise.stateChanged.connect(lambda state: self.updateCode())

    def initializePage(self):
        self.updateCode()

    def updateCode(self):
        perColumn = self.rbPerColumn.isChecked()

        if perColumn:
            columnNames = self.wizard().field("columnnames")
            usecols = self.wizard().field("usecols")

            if usecols is not None:  # User selected a subset of all columns
                # Pick corrsponding column names
                columnNames = [columnNames[i] for i in usecols]

            variables = ", ".join(columnNames)
        else:
            variables = "data"

        code = "import numpy\n"

        code += variables + " = numpy.genfromtxt(\n"
        for param, default in (
            ("fname", None),
            ("skip_header", 0),
            ("comments", "#"),
            ("delimiter", None),
            ("usecols", None),
            ("invalid_raise", True),
        ):
            value = self.wizard().field(param)
            if value != default:
                code += "\t%s = %r,\n" % (param, value)
        if perColumn:
            code += "\tunpack = True,\n"
        code += "\t)\n"

        self.codeView.setPlainText(code)

    def getCode(self):
        return self.codeView.toPlainText()

    def onBtnExecuteClicked(self):
        shell = pyzo.shells.getCurrentShell()
        if shell is None:
            QtWidgets.QMessageBox.information(
                self,
                translate("importwizard", "Import data wizard"),
                translate("importwizard", "No current shell active"),
            )
            return

        shell.executeCode(self.getCode(), "importwizard")

    def onBtnInsertClicked(self):
        editor = pyzo.editors.getCurrentEditor()
        if editor is None:
            QtWidgets.QMessageBox.information(
                self,
                translate("importwizard", "Import data wizard"),
                translate("importwizard", "No current file open"),
            )
            return

        code = self.getCode()

        # Format tabs/spaces according to editor setting
        if editor.indentUsingSpaces():
            code = code.replace("\t", " " * editor.indentWidth())

        # insert code at start of line
        cursor = editor.textCursor()
        cursor.movePosition(cursor.StartOfBlock)
        cursor.insertText(code)


class ImportWizard(QtWidgets.QWizard):
    def __init__(self):
        QtWidgets.QWizard.__init__(self)
        self.setMinimumSize(500, 400)
        self.resize(700, 500)

        self.setPreviewData(None)

        self.selectFilePage = SelectFilePage()
        self.setParametersPage = SetParametersPage()
        self.resultPage = ResultPage()

        self.addPage(self.selectFilePage)
        self.addPage(self.setParametersPage)
        self.addPage(self.resultPage)

        self.setWindowTitle(translate("importwizard", "Import data"))

        self.currentIdChanged.connect(self.onCurrentIdChanged)

    def onCurrentIdChanged(self, id):
        # Hide the 'cancel' button on the last page
        if self.nextId() == -1:
            self.button(QtWidgets.QWizard.CancelButton).hide()
        else:
            self.button(QtWidgets.QWizard.CancelButton).show()

    def open(self, filename):
        if self.isVisible():
            QtWidgets.QMessageBox.information(
                self,
                translate("importwizard", "Import data wizard"),
                translate("importwizard", "The import data wizard is already open"),
            )
            return

        self.restart()
        self.selectFilePage.txtFilename.setText(filename)
        self.selectFilePage.updatePreview()
        self.show()

    def field(self, name):
        # Allow access to all data via field, some properties are not avaialble
        # as actual controls and therefore we have to handle them ourselves
        if name == "usecols":
            return self.setParametersPage.selectedColumns()
        elif name == "columnnames":
            return self.setParametersPage.columnNames()
        else:
            return QtWidgets.QWizard.field(self, name)

    def setPreviewData(self, data):
        self._previewData = data

    def previewData(self):
        if self._previewData is None:
            raise RuntimeError("Preview data not loaded")

        return self._previewData


if __name__ == "__main__":
    iw = ImportWizard()
    iw.open("test.txt")



# -*- coding: utf-8 -*-
# flake8: noqa
# Copyright (C) 2013, the Pyzo development team
#
# Yoton is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

"""

The channel classes represent the mechanism for the user to send
messages into the network and receive messages from it. A channel
needs a context to function; the context represents a node in the
network.

"""

from yoton.channels.message_types import MessageType, TEXT, BINARY, OBJECT
from yoton.channels.channels_base import BaseChannel
from yoton.channels.channels_pubsub import PubChannel, SubChannel, select_sub_channel
from yoton.channels.channels_reqrep import (
    ReqChannel,
    RepChannel,
    Future,
    TimeoutError,
    CancelledError,
)
from yoton.channels.channels_state import StateChannel
from yoton.channels.channels_file import FileWrapper



# -*- coding: utf-8 -*-
# Copyright (C) 2013, the codeeditor development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

"""
Code editor extensions that change its behaviour (i.e. how it reacts to keys)
"""


from ..qt import QtGui, QtCore

Qt = QtCore.Qt

import pyzo
from ..misc import ustr, ce_option
from ..parsers.tokens import (
    CommentToken,
    IdentifierToken,
    StringToken,
    UnterminatedStringToken,
)
from ..parsers.python_parser import MultilineStringToken
from ..parsers import BlockState


class MoveLinesUpDown(object):
    def keyPressEvent(self, event):
        if event.key() in (Qt.Key_Up, Qt.Key_Down) and (
            Qt.ControlModifier & event.modifiers()
            and Qt.ShiftModifier & event.modifiers()
        ):

            cursor = self.textCursor()
            cursor.beginEditBlock()
            try:
                self._swaplines(cursor, event.key())
            finally:
                cursor.endEditBlock()

        else:
            super().keyPressEvent(event)

    def _swaplines(self, cursor, key):

        # Get positions of selection
        start = cursor.selectionStart()
        end = cursor.selectionEnd()

        # Get text of selected blocks
        cursor.setPosition(start, cursor.MoveAnchor)
        cursor.movePosition(cursor.StartOfBlock, cursor.MoveAnchor)
        cursor.setPosition(end, cursor.KeepAnchor)
        cursor.movePosition(cursor.EndOfBlock, cursor.KeepAnchor)
        text1 = cursor.selectedText()
        cursor.removeSelectedText()
        pos1 = cursor.position()

        # Move up/down
        other = [cursor.NextBlock, cursor.PreviousBlock][int(bool(key == Qt.Key_Up))]
        cursor.movePosition(other, cursor.MoveAnchor)

        # Select text of other block
        cursor.movePosition(cursor.StartOfBlock, cursor.MoveAnchor)
        cursor.movePosition(cursor.EndOfBlock, cursor.KeepAnchor)
        text2 = cursor.selectedText()
        cursor.removeSelectedText()
        pos2 = cursor.position()

        # Insert text
        cursor.insertText(text1)
        pos3 = cursor.position()

        # Move back
        if key == Qt.Key_Up:
            cursor.movePosition(cursor.NextBlock, cursor.MoveAnchor)
        else:
            cursor.setPosition(pos1, cursor.MoveAnchor)
            pos2 += len(text2)
            pos3 += len(text2)

        # Replace text
        cursor.insertText(text2)

        # Leave original lines selected for continued movement
        cursor.setPosition(pos2, cursor.MoveAnchor)
        cursor.setPosition(pos3, cursor.KeepAnchor)
        self.setTextCursor(cursor)


class ScrollWithUpDownKeys(object):
    def keyPressEvent(self, event):
        if (
            event.key() in (Qt.Key_Up, Qt.Key_Down)
            and Qt.ControlModifier == event.modifiers()
        ):
            s = self.verticalScrollBar()
            # h = self.cursorRect(self.textCursor()).height()
            if event.key() == Qt.Key_Up:
                s.setValue(s.value() + 1)
            else:
                s.setValue(s.value() - 1)

        else:
            super().keyPressEvent(event)


class HomeKey(object):
    def keyPressEvent(self, event):
        # Home or shift + home
        if event.key() == Qt.Key_Home and event.modifiers() in (
            Qt.NoModifier,
            Qt.ShiftModifier,
        ):
            # Prepare
            cursor = self.textCursor()
            shiftDown = event.modifiers() == Qt.ShiftModifier
            moveMode = [cursor.MoveAnchor, cursor.KeepAnchor][shiftDown]
            # Get leading whitespace
            text = ustr(cursor.block().text())
            leadingWhitespace = text[: len(text) - len(text.lstrip())]
            # Get current position and move to start of whitespace
            i = cursor.positionInBlock()
            cursor.movePosition(cursor.StartOfBlock, moveMode)
            cursor.movePosition(cursor.Right, moveMode, len(leadingWhitespace))
            # If we were alread there, move to start of block
            if cursor.positionInBlock() == i:
                cursor.movePosition(cursor.StartOfBlock, moveMode)
            # Done
            self.setTextCursor(cursor)
        else:
            super().keyPressEvent(event)


class EndKey(object):
    def keyPressEvent(self, event):
        if event.key() == Qt.Key_End and event.modifiers() in (
            Qt.NoModifier,
            Qt.ShiftModifier,
        ):
            # Prepare
            cursor = self.textCursor()
            shiftDown = event.modifiers() == Qt.ShiftModifier
            moveMode = [cursor.MoveAnchor, cursor.KeepAnchor][shiftDown]
            # Get current position and move to end of line
            i = cursor.positionInBlock()
            cursor.movePosition(cursor.EndOfLine, moveMode)
            # If alread at end of line, move to end of block
            if cursor.positionInBlock() == i:
                cursor.movePosition(cursor.EndOfBlock, moveMode)
            # Done
            self.setTextCursor(cursor)
        else:
            super().keyPressEvent(event)


class NumpadPeriodKey(object):
    """
    If the numpad decimal separator key is pressed, always insert
    a period (.) even if due to localization that key is mapped to a
    comma (,). When editing code, period is the decimal separator
    independent of localization
    """

    # NOTE: THIS PLUGIN IS DISABLED, see issue #720

    def keyPressEvent(self, event):
        # Check for numpad comma
        if (
            event.key() == QtCore.Qt.Key_Comma
            and event.modifiers() & QtCore.Qt.KeypadModifier
        ):

            # Create a new QKeyEvent to substitute the original one
            event = QtGui.QKeyEvent(
                event.type(),
                QtCore.Qt.Key_Period,
                event.modifiers(),
                ".",
                event.isAutoRepeat(),
                event.count(),
            )

        super().keyPressEvent(event)


class Indentation(object):
    def __cursorIsInLeadingWhitespace(self, cursor=None):
        """
        Checks wether the given cursor is in the leading whitespace of a block, i.e.
        before the first non-whitespace character. The cursor is not modified.
        If the cursor is not given or is None, the current textCursor is used
        """
        if cursor is None:
            cursor = self.textCursor()

        # Get the text of the current block up to the cursor
        textBeforeCursor = ustr(cursor.block().text())[: cursor.positionInBlock()]
        return (
            textBeforeCursor.lstrip() == ""
        )  # If we trim it and it is empty, it's all whitespace

    def keyPressEvent(self, event):
        key = event.key()
        modifiers = event.modifiers()
        # Tab key
        if key == Qt.Key_Tab:
            if modifiers == Qt.NoModifier:
                if (
                    self.textCursor().hasSelection()
                ):  # Tab pressed while some area was selected
                    self.indentSelection()
                    return
                elif self.__cursorIsInLeadingWhitespace():
                    # If the cursor is in the leading whitespace, indent and move cursor to end of whitespace
                    cursor = self.textCursor()
                    self.indentBlock(cursor)
                    self.setTextCursor(cursor)
                    return

                elif self.indentUsingSpaces():
                    # Insert space-tabs
                    cursor = self.textCursor()
                    w = self.indentWidth()
                    cursor.insertText(" " * (w - ((cursor.positionInBlock() + w) % w)))
                    return
                # else: default behaviour, insert tab character
            else:  # Some other modifiers + Tab: ignore
                return

        # If backspace is pressed in the leading whitespace, (except for at the first
        # position of the line), and there is no selection
        # dedent that line and move cursor to end of whitespace
        if (
            key == Qt.Key_Backspace
            and modifiers == Qt.NoModifier
            and self.__cursorIsInLeadingWhitespace()
            and not self.textCursor().atBlockStart()
            and not self.textCursor().hasSelection()
        ):
            # Create a cursor, dedent the block and move screen cursor to the end of the whitespace
            cursor = self.textCursor()
            self.dedentBlock(cursor)
            self.setTextCursor(cursor)
            return

        # todo: Same for delete, I think not (what to do with the cursor?)

        # Auto-unindent
        if key == Qt.Key_Delete:
            cursor = self.textCursor()
            if not cursor.hasSelection():
                cursor.movePosition(cursor.EndOfBlock, cursor.KeepAnchor)
                if not cursor.hasSelection() and cursor.block().next().isValid():
                    cursor.beginEditBlock()
                    cursor.movePosition(cursor.NextBlock)
                    self.indentBlock(cursor, -99)  # dedent as much as we can
                    cursor.deletePreviousChar()
                    cursor.endEditBlock()
                    return

        super().keyPressEvent(event)


class AutoIndent(object):
    """
    Auto indentation. This extension only adds the autoIndent property, for the
    actual indentation, the editor should derive from some AutoIndenter object
    """

    def autoIndent(self):
        """autoIndent()

        Get whether auto indentation is enabled.

        """
        return self.__autoIndent

    @ce_option(True)
    def setAutoIndent(self, value):
        """setAutoIndent(value)

        Set whether to enable auto indentation.

        """
        self.__autoIndent = bool(value)


class PythonAutoIndent(object):
    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        if not self.autoIndent():
            return

        # This extension code is run *after* key is processed by QPlainTextEdit

        if event.key() in (Qt.Key_Enter, Qt.Key_Return):
            cursor = self.textCursor()

            # Prevent in-block newlines (issue #482)
            if not cursor.atBlockStart() and not cursor.hasSelection():
                cursor.deletePreviousChar()
                cursor.insertBlock()
                cursor = self.textCursor()

            previousBlock = cursor.block().previous()
            if previousBlock.isValid():
                line = ustr(previousBlock.text())
                indent = line[: len(line) - len(line.lstrip())]
                if line.endswith(":"):
                    # We only need to add indent if the : is not in a (multiline)
                    # string or comment. Therefore, find out what the syntax
                    # highlighter thinks of the previous line.
                    ppreviousBlock = (
                        previousBlock.previous()
                    )  # the block before previous
                    ppreviousState = (
                        ppreviousBlock.userState() if previousBlock.isValid() else 0
                    )
                    lastElementToken = list(
                        self.parser().parseLine(previousBlock.text(), ppreviousState)
                    )[-1]
                    # Because there's at least a : on that line, the list is never empty

                    if not isinstance(
                        lastElementToken,
                        (CommentToken, UnterminatedStringToken, BlockState),
                    ):
                        # TODO: check correct identation (no mixed space/tabs)
                        if self.indentUsingSpaces():
                            indent += " " * self.indentWidth()
                        else:
                            indent += "\t"
                cursor.insertText(indent)
                # This prevents jump to start of line when up key is pressed
                self.setTextCursor(cursor)


class SmartCopyAndPaste(object):
    """
    Smart copy and paste allows copying and pasting blocks

    """

    @staticmethod
    def __setCursorPositionAndAnchor(cursor, position, anchor):
        cursor.setPosition(anchor)
        cursor.setPosition(position, cursor.KeepAnchor)

    @classmethod
    def __ensureCursorBeforeAnchor(cls, cursor):
        """
        Given a cursor, modify it such that the cursor.position() is before or
        at cursor.anchor() and not the other way around.

        Returns: anchorBeforeCursor, i.e. whether originally the anchor was
        before the cursor
        """
        start = cursor.selectionStart()
        end = cursor.selectionEnd()
        # Remember whether the cursor is before or after the anchor
        anchorBeforeCursor = cursor.anchor() < cursor.position()

        cls.__setCursorPositionAndAnchor(cursor, start, end)

        # Return wheter originally the cursor was before the anchor
        return anchorBeforeCursor

    def copy(self):
        """
        Smart copy: if selection is multi-line and in front of the start of the
        selection there is only whitespace, extend the selection to include only
        whitespace
        """
        cursor = self.textCursor()
        start = cursor.selectionStart()
        end = cursor.selectionEnd()

        # For our convenience, ensure that position is at start and
        # anchor is at the end, but remember whether originally the
        # anchor was before the cursor or the other way around
        anchorBeforeCursor = self.__ensureCursorBeforeAnchor(cursor)

        # Check if we have multi-line selection.
        block = cursor.block()
        # Use > not >= to ensure we don't count it as multi-line if the cursor
        # is just at the beginning of the next block (consistent with 'CodeEditor.doForSelectedLines')
        if end > (block.position() + block.length()):

            # Now check if there is only whitespace before the start of selection
            # If so, include this whitespace in the selection and update the
            # selection of the editor
            textBeforeSelection = block.text()[: cursor.positionInBlock()]
            if len(textBeforeSelection.strip()) == 0:
                start = block.position()  # Move start to include leading whitespace

                # Update the textcursor of our editor. If originally the
                # anchor was before the cursor, restore that situation
                if anchorBeforeCursor:
                    self.__setCursorPositionAndAnchor(cursor, end, start)
                else:
                    self.__setCursorPositionAndAnchor(cursor, start, end)

                self.setTextCursor(cursor)

        # Call our supers copy slot to do the actual copying
        super().copy()

    def cut(self):
        """
        Cutting with smart-copy: the part that is copies is the same as self.copy(),
        but the part that is removed is only the original selection

        see: Qt qtextcontrol.cpp, cut()
        """
        if (
            self.textInteractionFlags() & QtCore.Qt.TextEditable
        ) and self.textCursor().hasSelection():

            cursor = self.textCursor()
            self.copy()
            # Restore original cursor
            self.setTextCursor(cursor)
            cursor.removeSelectedText()

    def paste(self):
        """
        Smart paste
        If you paste on a position that has only whitespace in front of it,
        remove the whitespace before pasting. Combined with smart copy,
        this ensure indentation of the
        """
        self._paste(keepSelection=False)

    def pasteAndSelect(self):
        """
        Smart paste
        Like paste(), but keep the part that was pasted selected. This allows
        you to change the indentation after pasting using tab / shift-tab
        """
        self._paste(keepSelection=True)

    def _paste(self, keepSelection):
        # Create a cursor of the current selection
        cursor = self.textCursor()

        # Ensure that position is at start and anchor is at the end.
        # This is required to ensure that with keepPositiobOnInsert
        # set, the cursor will equal the pasted text after the pasting
        self.__ensureCursorBeforeAnchor(cursor)

        # Change this cursor to let the position() stay at its place upon
        # inserting the new code; the anchor will move to the end of the insertion
        cursor.setKeepPositionOnInsert(True)
        super().paste()

        block = cursor.block()

        # Check if the thing to be pasted is multi-line. Use > not >=
        # to ensure we don't count it as multi-line if the cursor
        # is just at the beginning of the next block (consistent with
        # 'CodeEditor.doForSelectedLines')
        if cursor.selectionEnd() > block.position() + block.length():

            # Now, check if in front of the current selection there is only whitespace
            if len(block.text()[: cursor.positionInBlock()].strip()) == 0:
                # Note that this 'smart pasting' will be a separate item on the
                # undo stack. This is intentional: the user can undo the 'smart paste'
                # without undoing the paste
                cursor2 = QtGui.QTextCursor(cursor)
                cursor2.setPosition(
                    cursor2.position()
                )  # put the anchor where the cursor is

                # Move cursor2 to beginning of the line (selecting the whitespace)
                # and remove the selection
                cursor2.movePosition(cursor2.StartOfBlock, cursor2.KeepAnchor)
                cursor2.removeSelectedText()

        # Set the textcursor of this editor to the just-pasted selection
        if keepSelection:
            cursor.setKeepPositionOnInsert(False)
            self.setTextCursor(cursor)


class AutoCloseQuotesAndBrackets(object):
    """
    Automatic insertion of quotes, parenthesis, braces and brackets

    """

    def _get_token_at_cursor(self, cursor=None, relpos=0):
        """Get token at the (current or given) cursor position. Can be None."""
        if cursor is None:
            cursor = self.textCursor()
        pos = cursor.positionInBlock() + relpos
        tokens = cursor.block().userData().tokens
        token = None
        for token in tokens:
            if hasattr(token, "start"):
                if token.start >= pos:
                    break
            elif getattr(token, "state", 0) in (1, 2):
                token = MultilineStringToken()  # 1 and 2 are mls, by convention, sortof
        return token

    def keyPressEvent(self, event):
        try:
            self.__keyPressEvent(event)
        except Exception as err:
            # When there is a bug in our fancy autoclosing stuff, better print and
            # and have the plain behavior instead of not working ...
            print(err)
            super().keyPressEvent(event)

    def __keyPressEvent(self, event):

        quotes = "'", '"'
        openBrackets = "{", "[", "("
        closeBrackets = "}", "]", ")"
        brackets = openBrackets + closeBrackets

        cursor = self.textCursor()
        char = event.text()

        #  brackets
        if char in brackets and pyzo.config.settings.autoClose_Brackets:

            # Dont autobracket inside comments and strings
            if isinstance(
                self._get_token_at_cursor(cursor),
                (
                    CommentToken,
                    StringToken,
                    MultilineStringToken,
                    UnterminatedStringToken,
                ),
            ):
                super().keyPressEvent(event)
                return

            if char in openBrackets:
                idx = openBrackets.index(char)
                next_char = self.__getNextCharacter()
                if cursor.selectedText():
                    # Surround selection with brackets
                    new_text = (
                        openBrackets[idx] + cursor.selectedText() + closeBrackets[idx]
                    )
                    cursor.setKeepPositionOnInsert(True)
                    cursor.insertText(new_text)
                    cursor.setKeepPositionOnInsert(False)
                    self.setTextCursor(cursor)
                elif (
                    next_char.strip() and next_char not in closeBrackets
                ):  # str.strip() conveniently removes all kinds of whitespace
                    # Only autoclose if the char on the right is whitespace
                    cursor.insertText(char)  # == super().keyPressEvent(event)
                else:
                    # Auto-close bracket
                    insert_txt = "{}{}".format(openBrackets[idx], closeBrackets[idx])
                    cursor.insertText(insert_txt)
                    self._moveCursorLeft(1)

            elif char in closeBrackets:
                idx = closeBrackets.index(char)
                next_char = self.__getNextCharacter()
                if cursor.selectedText():
                    # Replace
                    cursor.insertText(char)
                elif next_char == char:
                    # Skip
                    self._moveCursorRight(1)
                else:
                    # Normal
                    cursor.insertText(char)  # == super().keyPressEvent(event)

        # quotes
        elif char in quotes and pyzo.config.settings.autoClose_Quotes:
            next_char = self.__getNextCharacter()

            # Dont autoquote inside comments and multiline strings
            # Only allow "doing our thing" when we're at the end of a normal string
            token = self._get_token_at_cursor(cursor)
            if isinstance(token, (CommentToken, MultilineStringToken)):
                super().keyPressEvent(event)
                return
            elif isinstance(token, StringToken) and char != next_char:
                super().keyPressEvent(event)
                return

            if cursor.selectedText() in ('"', "'"):
                # Skip over char if its one char and a quote
                self._moveCursorRight(1)
            elif cursor.selectedText():
                # Surround selection with quotes, maybe even multi-line
                new_text = char + cursor.selectedText() + char
                if "\u2029" in new_text and "python" in self.parser().name().lower():
                    new_text = char * 2 + new_text + char * 2
                cursor.setKeepPositionOnInsert(True)
                cursor.insertText(new_text)
                cursor.setKeepPositionOnInsert(False)
                self.setTextCursor(cursor)
            elif next_char == char:
                # Skip
                self._moveCursorRight(1)
            else:
                # Only autoquote if we're next to whitespace, operator, quote
                notok_token_types = (IdentifierToken,)
                tokenL = self._get_token_at_cursor(cursor, -1)
                tokenR = self._get_token_at_cursor(cursor, -0)
                if isinstance(tokenL, notok_token_types) or isinstance(
                    tokenR, notok_token_types
                ):
                    super().keyPressEvent(event)
                    return
                # Auto-close
                cursor.insertText(char * 2)
                self._moveCursorLeft(1)
                # # Maybe handle tripple quotes (add 2 more if we now have 3 on the left)
                # Disabled: this feature too easily gets in the way
                # if 'python' in self.parser().name().lower():
                #     cursor = self.textCursor()
                #     cursor.movePosition(cursor.PreviousCharacter, cursor.KeepAnchor, 3)
                #     if cursor.selectedText() == char * 3:
                #         edit_cursor = self.textCursor()
                #         edit_cursor.insertText(char * 2)
                #         self._moveCursorLeft(2)

        # remove whole couple of brackets when hitting backspace
        elif (
            event.key() == Qt.Key_Backspace and pyzo.config.settings.autoClose_Brackets
        ):
            if isinstance(
                self._get_token_at_cursor(cursor),
                (
                    CommentToken,
                    StringToken,
                    MultilineStringToken,
                    UnterminatedStringToken,
                ),
            ):
                super().keyPressEvent(event)
                return

            else:
                prev_char = self.__getPreviousCharacter()
                next_char = self.__getNextCharacter()

                if prev_char == "(" and next_char == ")":
                    cursor.deleteChar()
                elif prev_char == "[" and next_char == "]":
                    cursor.deleteChar()
                elif prev_char == "{" and next_char == "}":
                    cursor.deleteChar()

                super().keyPressEvent(event)

        else:
            super().keyPressEvent(event)

    def __getNextCharacter(self):
        cursor = self.textCursor()
        cursor.movePosition(cursor.NoMove, cursor.MoveAnchor)  # rid selection
        cursor.movePosition(cursor.NextCharacter, cursor.KeepAnchor)
        next_char = cursor.selectedText()
        return next_char

    def __getPreviousCharacter(self):
        cursor = self.textCursor()
        cursor.movePosition(cursor.NoMove, cursor.MoveAnchor)  # rid selection
        cursor.movePosition(cursor.PreviousCharacter, cursor.KeepAnchor)
        previous_char = cursor.selectedText()
        return previous_char

    def _moveCursorLeft(self, n):
        """
        Move cursor left between eg. brackets
        """
        cursor2 = self.textCursor()
        cursor2.movePosition(cursor2.Left, cursor2.MoveAnchor, n)
        self.setTextCursor(cursor2)

    def _moveCursorRight(self, n):
        """
        Move cursor out of eg. brackets
        """
        cursor2 = self.textCursor()
        cursor2.movePosition(cursor2.Right, cursor2.MoveAnchor, n)
        self.setTextCursor(cursor2)



""" This is a bit awkward, but yoton is a package that is designed to
work from Python 2.4 to Python 3.x. As such, it does not have relative
imports and must be imported as an absolute package. That is what this
module does...
"""

import os
import sys

# Allow importing yoton
sys.path.insert(0, os.path.dirname(__file__))

# Import yoton
import yoton  # noqa

# Reset
sys.path.pop(0)



# -*- coding: utf-8 -*-
# Copyright (C) 2014, the Pyzo development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

""" Module to deal with command line arguments.

In specific, this allows doing "pyzo some_file.py" and the file will be
opened in an existing pyzo window (if available) or a new pyzo process
is started to open the file.

This module is used at the very early stages of starting pyzo, and also
in main.py to apply any command line args for the current process, and
to closse down the server when pyzo is closed.
"""

import sys
import os

from yoton.clientserver import RequestServer, do_request
import pyzo

# Local address to host on. we use yoton's port hash to have an arbitrary port
ADDRESS = "localhost:pyzoserver"


class Server(RequestServer):
    """Server that listens on a port for commands.
    The commands can be send by executing the Pyzo executable with
    command line arguments.
    """

    def handle_request(self, request):
        """This is where the requests enter."""
        # Get command
        request = request.strip()
        command, _, arg = request.partition(" ")

        # Handle command
        try:
            reply = handle_command(command, arg)
        except Exception as err:
            msg = "Error handling request %r:\n%s" % (request, str(err))
            pyzo.callLater(print, msg)
            return msg
        else:
            pyzo.callLater(print, "Request:", request)
            pyzo.callLater(print, "Reply:", reply)
            return reply


def handle_command(command, arg):
    """Function that handles all pyzo commands.
    This gets called either from the server, or from the code that
    processed command line args.
    """
    if not command:
        return "empty command?"

    elif command == "testerr":
        return 1 / 0

    elif command == "stopserver":
        # For efficiently stopping the server
        if server:
            server.stop()
            return "Stopped the server"

    elif command == "echo":
        # For testing
        return "echo %r" % arg

    elif command == "open":
        # Open a file in the editor
        if not arg:
            return "The open command requires a filename."
        pyzo.callLater(pyzo.editors.loadFile, arg)
        return "Opened file %r" % arg

    elif command == "new":
        # Open a new (temp) file in the editor
        pyzo.callLater(pyzo.editors.newFile)
        return "Created new file"

    elif command == "close":
        # Close pyzo
        pyzo.callLater(pyzo.main.close)
        return "Closing Pyzo"

    else:
        # Assume the user wanted to open a file
        fname = (command + " " + arg).rstrip()
        if not pyzo.editors:
            return "Still warming up ..."
        else:
            pyzo.callLater(pyzo.editors.loadFile, fname)
            return "Try opening file %r" % fname

    # We should always return. So if we get here, it is a bug.
    # Return something so that we can be aware.
    return "error " + command


def handle_cmd_args():
    """Handle command line arguments by sending them to the server.
    Returns a result string if any commands were processed, and None
    otherwise.
    """
    args = sys.argv[1:]
    request = " ".join(arg for arg in args if not arg.startswith("--"))
    if "psn_" in request and not os.path.isfile(request):
        request = " ".join(args[1:])  # An OSX thing when clicking app icon
    request = request.strip()
    #
    if not request:
        return None
    else:
        # Always send to server, even if we are the ones that run the server
        try:
            return do_request(ADDRESS, request, 0.4).rstrip()
        except Exception as err:
            print("Could not process command line args:\n%s" % str(err))
            return None


def stop_our_server():
    """Stop our server, for shutting down nicely.
    This is faster than calling server.stop(), because in the latter
    case the server will need to timeout (0.25 s) before it sees that
    it needs to stop.
    """
    if is_our_server_running():
        try:
            server.stop()  # Post a stop message
            do_request(ADDRESS, "stopserver", 0.1)  # trigger
            print("Stopped our command server.")
        except Exception as err:
            print("Failed to stop command server:")
            print(err)


def is_our_server_running():
    """Return True if our server is running. If it is, this process
    is the main Pyzo; the first Pyzo that was started. If the server is
    not running, this is probably not the first Pyzo, but there might
    also be problem with starting the server.
    """
    return server and server.is_alive()


def is_pyzo_server_running():
    """Test whether the Pyzo server is running *somewhere* (not
    necesarily in this process).
    """
    try:
        res = do_request(ADDRESS, "echo", 0.2)
        return res.startswith("echo")
    except Exception:
        return False


# Shold we start the server?
_try_start_server = True
if sys.platform.startswith("win"):
    _try_start_server = not is_pyzo_server_running()


# Create server
server_err = None
server = None
try:
    if _try_start_server:
        server = Server(ADDRESS)
        server.start()
except OSError as err:
    server_err = err
    server = None



# -*- coding: utf-8 -*-
# Copyright (C) 2013, the Pyzo development team
#
# Yoton is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

""" Module yoton.channels.channels_pubsub

Defines the channel classes for the pub/sub pattern.

"""

import time

from yoton.misc import bytes, xrange
from yoton.channels import BaseChannel

QUEUE_NULL = 0
QUEUE_OK = 1
QUEUE_FULL = 2


class PubChannel(BaseChannel):
    """PubChannel(context, slot_base, message_type=yoton.TEXT)

    The publish part of the publish/subscribe messaging pattern.
    Sent messages are received by all yoton.SubChannel instances with
    the same slot.

    There are no limitations for this channel if events are not processed.

    Parameters
    ----------
    context : yoton.Context instance
        The context that this channel uses to send messages in a network.
    slot_base : string
        The base slot name. The channel appends an extension to indicate
        message type and messaging pattern to create the final slot name.
        The final slot is used to connect channels at different contexts
        in a network
    message_type : yoton.MessageType instance
        (default is yoton.TEXT)
        Object to convert messages to bytes and bytes to messages.
        Users can create their own message_type class to let channels
        any type of message they want.

    """

    def __init__(self, *args, **kwargs):
        BaseChannel.__init__(self, *args, **kwargs)
        self._source_set = set()

    def _messaging_patterns(self):
        return "pub-sub", "sub-pub"

    def send(self, message):
        """send(message)

        Send a message over the channel. What is send as one
        message will also be received as one message.

        The message is queued and delivered to all corresponding
        SubChannels (i.e. with the same slot) in the network.

        """
        self._send(self.message_to_bytes(message))

    def _recv_package(self, package):
        """Overloaded to set blocking mode.
        Do not call _maybe_emit_received(), a PubChannel never emits
        the "received" signal.
        """

        message = package._data.decode("utf-8")
        source_id = package._source_id

        # Keep track of who's queues are full
        if message == "full":
            self._source_set.add(source_id)
        else:
            self._source_set.discard(source_id)

        # Set lock if there is a channel with a full queue,
        # Unset if there are none
        if self._source_set:
            self._set_send_lock(True)
            # sys.stderr.write('setting lock\n')
        else:
            self._set_send_lock(False)
            # sys.stderr.write('unsetting lock\n')


class SubChannel(BaseChannel):
    """SubChannel(context, slot_base, message_type=yoton.TEXT)

    The subscribe part of the publish/subscribe messaging pattern.
    Received messages were sent by a yoton.PubChannel instance at the
    same slot.

    This channel can be used as an iterator, which yields all pending
    messages. The function yoton.select_sub_channel can
    be used to synchronize multiple SubChannel instances.

    If no events being processed this channel works as normal, except
    that the received signal will not be emitted, and sync mode will
    not work.

    Parameters
    ----------
    context : yoton.Context instance
        The context that this channel uses to send messages in a network.
    slot_base : string
        The base slot name. The channel appends an extension to indicate
        message type and messaging pattern to create the final slot name.
        The final slot is used to connect channels at different contexts
        in a network
    message_type : yoton.MessageType instance
        (default is yoton.TEXT)
        Object to convert messages to bytes and bytes to messages.
        Users can create their own message_type class to let channels
        any type of message they want.

    """

    def __init__(self, *args, **kwargs):
        BaseChannel.__init__(self, *args, **kwargs)

        # To detect when to block the sending side
        self._queue_status = QUEUE_NULL
        self._queue_status_timeout = 0
        self._HWM = 32
        self._LWM = 16

        # Automatically check queue status when new data
        # enters the system
        self.received.bind(self._check_queue_status)

    def _messaging_patterns(self):
        return "sub-pub", "pub-sub"

    def __iter__(self):
        return self

    def __next__(self):  # Python 3.x
        m = self.recv(False)
        if m:
            return m
        else:
            raise StopIteration()

    def next(self):  # Python 2.x
        """next()

        Return the next message, or raises StopIteration if non available.

        """
        return self.__next__()

    ## For sync mode

    def set_sync_mode(self, value):
        """set_sync_mode(value)

        Set or unset the SubChannel in sync mode. When in sync mode, all
        channels that send messages to this channel are blocked if
        the queue for this SubChannel reaches a certain size.

        This feature can be used to limit the rate of senders if the consumer
        (i.e. the one that calls recv()) cannot keep up with processing
        the data.

        This feature requires the yoton event loop to run at the side
        of the SubChannel (not necessary for the yoton.PubChannel side).

        """
        value = bool(value)

        # First reset block status if necessary
        if self._queue_status == QUEUE_FULL:
            self._send_block_message_to_senders("ok")

        # Set new queue status flag
        if value:
            self._queue_status = QUEUE_OK
        else:
            self._queue_status = QUEUE_NULL

    def _send_block_message_to_senders(self, what):
        """_send_block_message_to_senders(what)

        Send a message to the PubChannel side to make it block/unblock.

        """

        # Check
        if not self._context.connection_count:
            return

        # Send
        try:
            self._send(what.encode("utf-8"))
        except IOError:
            # If self._closed
            self._check_queue_status = QUEUE_NULL

    def _check_queue_status(self, dummy=None):
        """_check_queue_status()

        Check the queue status. Returns immediately unless this receiving
        channel runs in sync mode.

        If the queue is above a certain size, will send out a package that
        will make the sending side block. If the queue is below a certain
        size, will send out a package that will make the sending side unblock.

        """

        if self._queue_status == QUEUE_NULL:
            return
        elif len(self._q_in) > self._HWM:
            if self._queue_status == QUEUE_OK:
                self._queue_status = QUEUE_FULL
                self._queue_status_timeout = time.time() + 4.0
                self._send_block_message_to_senders("full")
        elif len(self._q_in) < self._LWM:
            if self._queue_status == QUEUE_FULL:
                self._queue_status = QUEUE_OK
                self._queue_status_timeout = time.time() + 4.0
                self._send_block_message_to_senders("ok")

        # Resend every so often. After 10s the PubChannel will unlock itself
        if self._queue_status_timeout < time.time():
            self._queue_status_timeout = time.time() + 4.0
            if self._queue_status == QUEUE_OK:
                self._send_block_message_to_senders("ok")
            else:
                self._send_block_message_to_senders("full")

    ## Receive methods

    def recv(self, block=True):
        """recv(block=True)

        Receive a message from the channel. What was send as one
        message is also received as one message.

        If block is False, returns empty message if no data is available.
        If block is True, waits forever until data is available.
        If block is an int or float, waits that many seconds.
        If the channel is closed, returns empty message.

        """

        # Check queue status, maybe we need to block the sender
        self._check_queue_status()

        # Get package
        package = self._recv(block)

        # Return message content or None
        if package is not None:
            return self.message_from_bytes(package._data)
        else:
            return self.message_from_bytes(bytes())

    def recv_all(self):
        """recv_all()

        Receive a list of all pending messages. The list can be empty.

        """

        # Check queue status, maybe we need to block the sender
        self._check_queue_status()

        # Pop all messages and return as a list
        pop = self._q_in.pop
        packages = [pop() for i in xrange(len(self._q_in))]
        return [self.message_from_bytes(p._data) for p in packages]

    def recv_selected(self):
        """recv_selected()

        Receive a list of messages. Use only after calling
        yoton.select_sub_channel with this channel as one of the arguments.

        The returned messages are all received before the first pending
        message in the other SUB-channels given to select_sub_channel.

        The combination of this method and the function select_sub_channel
        enables users to combine multiple SUB-channels in a way that
        preserves the original order of the messages.

        """

        # No need to check queue status, we've done that in the
        # _get_pending_sequence_numbers() method

        # Prepare
        q = self._q_in
        ref_seq = self._ref_seq
        popped = []

        # Pop all messages that have sequence number lower than reference
        try:
            for i in xrange(len(q)):
                part = q.pop()
                if part._recv_seq > ref_seq:
                    q.insert(part)  # put back in queue
                    break
                else:
                    popped.append(part)
        except IndexError:
            pass

        # Done; return messages
        return [self.message_from_bytes(p._data) for p in popped]

    def _get_pending_sequence_numbers(self):
        """_get_pending_sequence_numbers()

        Get the sequence numbers of the first and last pending messages.
        Returns (-1,-1) if no messages are pending.

        Used by select_sub_channel() to determine which channel should
        be read from first and what the reference sequence number is.

        """

        # Check queue status, maybe we need to block the sender
        self._check_queue_status()

        # Peek
        try:
            q = self._q_in
            return q.peek(0)._recv_seq, q.peek(-1)._recv_seq + 1
        except IndexError:
            return -1, -1


def select_sub_channel(*args):
    """select_sub_channel(channel1, channel2, ...)

    Returns the channel that has the oldest pending message of all
    given yoton.SubCannel instances. Returns None if there are no pending
    messages.

    This function can be used to read from SubCannels instances in the
    order that the messages were send.

    After calling this function, use channel.recv_selected() to obtain
    all messages that are older than any pending messages in the other
    given channels.

    """

    # Init
    smallest_seq1 = 99999999999999999999999999
    smallest_seq2 = 99999999999999999999999999
    first_channel = None

    # For each channel ...
    for channel in args:

        # Check if channel is of right type
        if not isinstance(channel, SubChannel):
            raise ValueError("select_sub_channel() only accepts SUB channels.")

        # Get and check sequence
        seq1, seq2 = channel._get_pending_sequence_numbers()
        if seq1 >= 0:
            if seq1 < smallest_seq1:
                # Cannot go beyond number of packages in queue,
                # or than seq1 of earlier selected channel.
                smallest_seq2 = min(smallest_seq1, smallest_seq2, seq2)
                # Store
                smallest_seq1 = seq1
                first_channel = channel
            else:
                # The first_channel cannot go beyond the 1st package in THIS queue
                smallest_seq2 = min(smallest_seq2, seq1)

    # Set flag at channel and return
    if first_channel:
        first_channel._ref_seq = smallest_seq2
        return first_channel
    else:
        return None



# -*- coding: utf-8 -*-
# Copyright (c) 2016, Almar Klein

"""
This module implements functionality to detect available Python
interpreters. This is done by looking at common locations, the Windows
registry, and conda's environment list.
"""

import sys
import os

from .pythoninterpreter import EXE_DIR, PythonInterpreter, versionStringToTuple
from .inwinreg import get_interpreters_in_reg
from .. import paths


def get_interpreters(minimumVersion=None):
    """get_interpreters(minimumVersion=None)
    Returns a list of PythonInterpreter instances.
    If minimumVersion is given, return only the interprers with at least that
    version, and also sort the result by version number.
    """

    # Get Python interpreters
    if sys.platform.startswith("win"):
        pythons = _get_interpreters_win()
    else:
        pythons = _get_interpreters_posix()
    pythons = set([PythonInterpreter(p) for p in pythons])

    # Get conda paths
    condas = set([PythonInterpreter(p) for p in _get_interpreters_conda()])

    # Get relative interpreters
    relative = set([PythonInterpreter(p) for p in _get_interpreters_relative()])

    # Get Pyzo paths
    pyzos = set([PythonInterpreter(p) for p in _get_interpreters_pyzo()])

    # Get pipenv paths
    pipenvs = set([PythonInterpreter(p) for p in _get_interpreters_pipenv()])

    # Almost done
    interpreters = set.union(pythons, condas, relative, pyzos, pipenvs)
    minimumVersion = minimumVersion or "0"
    return _select_interpreters(interpreters, minimumVersion)


def _select_interpreters(interpreters, minimumVersion):
    """Given a list of PythonInterpreter instances, return a list with
    the interpreters selected that are valid and have their version equal
    or larger than the given minimimVersion. The returned list is sorted
    by version number.
    """
    if not isinstance(minimumVersion, str):
        raise ValueError("minimumVersion in get_interpreters must be a string.")
    # Remove invalid interpreters
    interpreters = [i for i in interpreters if i.version]
    # Remove the ones below the reference version
    if minimumVersion is not None:
        refTuple = versionStringToTuple(minimumVersion)
        interpreters = [i for i in interpreters if (i.version_info >= refTuple)]
    # Return, sorted by version
    return sorted(interpreters, key=lambda x: x.version_info)


def _get_interpreters_win():
    found = []

    # Query from registry
    for v in get_interpreters_in_reg():
        found.append(v.installPath())

    # Check common locations
    for rootname in [
        "C:/",
        "~/",
        "C:/program files/",
        "C:/program files (x86)/",
        "C:/ProgramData/",
        "~/appdata/local/programs/python/",
        "~/appdata/local/continuum/",
        "~/appdata/local/anaconda/",
    ]:
        rootname = os.path.expanduser(rootname)
        if not os.path.isdir(rootname):
            continue
        for dname in os.listdir(rootname):
            if dname.lower().startswith(("python", "pypy", "miniconda", "anaconda")):
                found.append(os.path.join(rootname, dname))

    # remove 'None' path added in the found list (when badly registered python in the user system)
    found = [v for v in found if v]

    # Normalize all paths, and remove trailing backslashes
    found = [os.path.normcase(os.path.abspath(v)).strip("\\") for v in found]

    # Append "python.exe" and check if that file exists
    found2 = []
    for dname in found:
        for fname in ("python.exe", "pypy.exe"):
            exename = os.path.join(dname, fname)
            if os.path.isfile(exename):
                found2.append(exename)
                break

    # Return as set (remove duplicates)
    return set(found2)


def _get_interpreters_posix():
    found = []

    # Look for system Python interpreters
    for searchpath in ["/usr/bin", "/usr/local/bin", "/opt/local/bin"]:
        searchpath = os.path.expanduser(searchpath)

        # Get files
        try:
            files = os.listdir(searchpath)
        except Exception:
            continue

        # Search for python executables
        for fname in files:
            if fname.startswith(("python", "pypy")) and not fname.count("config"):
                if len(fname) < 16:
                    # Get filename and resolve symlink
                    filename = os.path.join(searchpath, fname)
                    filename = os.path.realpath(filename)
                    # Seen on OS X that was not a valid file
                    if os.path.isfile(filename):
                        found.append(filename)

    # Look for user-installed Python interpreters such as pypy and anaconda
    for rootname in ["~", "/usr/local"]:
        rootname = os.path.expanduser(rootname)
        if not os.path.isdir(rootname):
            continue
        for dname in os.listdir(rootname):
            if dname.lower().startswith(("python", "pypy", "miniconda", "anaconda")):
                for fname in ("bin/python", "bin/pypy"):
                    exename = os.path.join(rootname, dname, fname)
                    if os.path.isfile(exename):
                        found.append(exename)

    # Remove pythonw, pythonm and the like
    found = set(found)
    for path in list(found):
        if path.endswith(("m", "w")) and path[:-1] in found:
            found.discard(path)

    # Return as set (remove duplicates)
    return set(found)


def _get_interpreters_pyzo():
    """Get a list of known Pyzo interpreters."""
    pythonname = "python" + ".exe" * sys.platform.startswith("win")
    exes = []
    for d in paths.pyzo_dirs():
        for fname in [
            os.path.join(d, "bin", pythonname + "3"),
            os.path.join(d, pythonname),
        ]:
            if os.path.isfile(fname):
                exes.append(fname)
                break
    return exes


def _get_interpreters_conda():
    """Get known conda environments"""
    if sys.platform.startswith("win"):
        pythonname = "python" + ".exe"
    else:
        pythonname = "bin/python"

    exes = []
    filename = os.path.expanduser("~/.conda/environments.txt")
    if os.path.isfile(filename):
        for line in open(filename, "rt").readlines():
            line = line.strip()
            exe_filename = os.path.join(line, pythonname)
            if line and os.path.isfile(exe_filename):
                exes.append(exe_filename)
    return exes


def _get_interpreters_pipenv():
    """Get known pipenv environments"""
    if sys.platform.startswith("win"):
        pythonname = "Scripts\\python" + ".exe"
    else:
        pythonname = "bin/python"

    exes = []
    for basedir in [os.path.join(os.path.expanduser("~"), ".virtualenvs")]:
        if os.path.isdir(basedir):
            for d in os.listdir(basedir):
                exe_filename = os.path.join(basedir, d, pythonname)
                if os.path.isfile(exe_filename):
                    exes.append(exe_filename)
    return exes


def _get_interpreters_relative():
    """Get interpreters relative to our prefix and the parent directory.
    This allows Pyzo to be shipped inside a pre-installed conda env.
    """
    pythonname = "python" + ".exe" * sys.platform.startswith("win")
    exes = []
    for d in [".", ".."]:
        for fname in [
            os.path.join(d, "bin", pythonname),
            os.path.join(d, pythonname),
        ]:
            filename = os.path.abspath(os.path.join(EXE_DIR, fname))
            if os.path.isfile(filename):
                exes.append(fname)
                break
    return exes


if __name__ == "__main__":
    for pi in get_interpreters():
        print(pi)



# -*- coding: utf-8 -*-
# Copyright (C) 2013, the Pyzo development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

""" pyzowizard module

Implements a wizard to help new users get familiar with pyzo.

"""

import os
import re

import pyzo
from pyzo.qt import QtCore, QtGui, QtWidgets
from pyzo import translate

from pyzo.util._locale import LANGUAGES, LANGUAGE_SYNONYMS, setLanguage


def retranslate(t):
    """To allow retranslating after selecting the language."""
    if hasattr(t, "original"):
        return translate("wizard", t.original)
    else:
        return t


class PyzoWizard(QtWidgets.QWizard):
    def __init__(self, parent):
        QtWidgets.QWizard.__init__(self, parent)

        # Set some appearance stuff
        self.setMinimumSize(600, 500)
        self.setWindowTitle(translate("wizard", "Getting started with Pyzo"))
        self.setWizardStyle(self.ModernStyle)
        self.setButtonText(self.CancelButton, "Stop")

        # Set logo
        pm = QtGui.QPixmap()
        pm.load(os.path.join(pyzo.pyzoDir, "resources", "appicons", "pyzologo48.png"))
        self.setPixmap(self.LogoPixmap, pm)

        # Define pages
        klasses = [
            IntroWizardPage,
            TwocomponentsWizardPage,
            EditorWizardPage,
            ShellWizardPage1,
            ShellWizardPage2,
            RuncodeWizardPage1,
            RuncodeWizardPage2,
            ToolsWizardPage1,
            ToolsWizardPage2,
            FinalPage,
        ]

        # Create pages
        self._n = len(klasses)
        for i, klass in enumerate(klasses):
            self.addPage(klass(self, i))

    def show(self, startPage=None):
        """Show the wizard. If startPage is given, open the Wizard at
        that page. startPage can be an integer or a string that matches
        the classname of a page.
        """
        QtWidgets.QWizard.show(self)

        # Check startpage
        if isinstance(startPage, int):
            pass
        elif isinstance(startPage, str):
            for i in range(self._n):
                page = self.page(i)
                if page.__class__.__name__.lower() == startPage.lower():
                    startPage = i
                    break
            else:
                print("Pyzo wizard: Could not find start page: %r" % startPage)
                startPage = None
        elif startPage is not None:
            print("Pyzo wizard: invalid start page: %r" % startPage)
            startPage = None

        # Go to start page
        if startPage is not None:
            for i in range(startPage):
                self.next()


class BasePyzoWizardPage(QtWidgets.QWizardPage):

    _prefix = translate("wizard", "Step")

    _title = "dummy title"
    _descriptions = []
    _image_filename = ""

    def __init__(self, parent, i):
        QtWidgets.QWizardPage.__init__(self, parent)
        self._i = i

        # Create label for description
        self._text_label = QtWidgets.QLabel(self)
        self._text_label.setTextFormat(QtCore.Qt.RichText)
        self._text_label.setWordWrap(True)

        # Create label for image
        self._comicLabel = QtWidgets.QLabel(self)
        pm = QtGui.QPixmap()
        if "logo" in self._image_filename:
            pm.load(
                os.path.join(
                    pyzo.pyzoDir, "resources", "appicons", self._image_filename
                )
            )
        elif self._image_filename:
            pm.load(
                os.path.join(pyzo.pyzoDir, "resources", "images", self._image_filename)
            )
        self._comicLabel.setPixmap(pm)
        self._comicLabel.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)

        # Layout
        theLayout = QtWidgets.QVBoxLayout(self)
        self.setLayout(theLayout)
        #
        theLayout.addWidget(self._text_label)
        theLayout.addStretch()
        theLayout.addWidget(self._comicLabel)
        theLayout.addStretch()

    def initializePage(self):

        # Get prefix
        i = self._i
        n = self.wizard()._n - 2  # Dont count the first and last page
        prefix = ""
        if i and i <= n:
            prefix = retranslate(self._prefix) + " %i/%i: " % (i, n)

        # Set title
        self.setTitle(prefix + retranslate(self._title))

        # Parse description
        # Two description units are separated with BR tags
        # Emphasis on words is set to italic tags.
        lines = []
        descriptions = [retranslate(d).strip() for d in self._descriptions]
        for description in descriptions:
            for line in description.splitlines():
                line = line.strip()
                line = re.sub(r"\*(.+?)\*", r"<b>\1</b>", line)
                lines.append(line)
            lines.append("<br /><br />")
        lines = lines[:-1]

        # Set description
        self._text_label.setText("\n".join(lines))


class IntroWizardPage(BasePyzoWizardPage):

    _title = translate("wizard", "Welcome to the Interactive Editor for Python!")
    _image_filename = "pyzologo128.png"
    _descriptions = [
        translate(
            "wizard",
            """This wizard helps you get familiarized with the workings of Pyzo.""",
        ),
        translate(
            "wizard",
            """Pyzo is a cross-platform Python IDE
        focused on *interactivity* and *introspection*, which makes it
        very suitable for scientific computing. Its practical design
        is aimed at *simplicity* and *efficiency*.""",
        ),
    ]

    def __init__(self, parent, i):
        BasePyzoWizardPage.__init__(self, parent, i)

        # Create label and checkbox
        t1 = translate("wizard", "This wizard can be opened using 'Help > Pyzo wizard'")
        # t2 = translate('wizard', "Show this wizard on startup")
        self._label_info = QtWidgets.QLabel(t1, self)
        # self._check_show = QtWidgets.QCheckBox(t2, self)
        # self._check_show.stateChanged.connect(self._setNewUser)

        # Create language switcher
        self._langLabel = QtWidgets.QLabel(translate("wizard", "Select language"), self)
        #
        self._langBox = QtWidgets.QComboBox(self)
        self._langBox.setEditable(False)
        # Fill
        index, theIndex = -1, -1
        cur = pyzo.config.settings.language
        for lang in sorted(LANGUAGES):
            index += 1
            self._langBox.addItem(lang)
            if lang == LANGUAGE_SYNONYMS.get(cur, cur):
                theIndex = index
        # Set current index
        if theIndex >= 0:
            self._langBox.setCurrentIndex(theIndex)
        # Bind signal
        self._langBox.activated.connect(self.onLanguageChange)

        # Init check state
        # if pyzo.config.state.newUser:
        #    self._check_show.setCheckState(QtCore.Qt.Checked)

        # Create sublayout
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self._langLabel, 0)
        layout.addWidget(self._langBox, 0)
        layout.addStretch(2)
        self.layout().addLayout(layout)

        # Add to layout
        self.layout().addSpacing(10)
        self.layout().addWidget(self._label_info)
        # self.layout().addWidget(self._check_show)

    def _setNewUser(self, newUser):
        newUser = bool(newUser)
        self._label_info.setHidden(newUser)
        pyzo.config.state.newUser = newUser

    def onLanguageChange(self):
        languageName = self._langBox.currentText()
        if pyzo.config.settings.language == languageName:
            return
        # Save new language
        pyzo.config.settings.language = languageName
        setLanguage(pyzo.config.settings.language)
        # Notify user
        text = translate(
            "wizard",
            """
        The language has been changed for this wizard.
        Pyzo needs to restart for the change to take effect application-wide.
        """,
        )
        m = QtWidgets.QMessageBox(self)
        m.setWindowTitle(translate("wizard", "Language changed"))
        m.setText(text)
        m.setIcon(m.Information)
        m.exec_()

        # Get props of current wizard
        geo = self.wizard().geometry()
        parent = self.wizard().parent()
        # Close ourself!
        self.wizard().close()
        # Start new one
        w = PyzoWizard(parent)
        w.setGeometry(geo)
        w.show()


class TwocomponentsWizardPage(BasePyzoWizardPage):

    _title = translate("wizard", "Pyzo consists of two main components")
    _image_filename = "pyzo_two_components.png"
    _descriptions = [
        translate("wizard", "You can execute commands directly in the *shell*,"),
        translate("wizard", "or you can write code in the *editor* and execute that."),
    ]


class EditorWizardPage(BasePyzoWizardPage):

    _title = translate("wizard", "The editor is where you write your code")
    _image_filename = "pyzo_editor.png"
    _descriptions = [
        translate(
            "wizard",
            """In the *editor*, each open file is represented as a tab. By
        right-clicking on a tab, files can be run, saved, closed, etc.""",
        ),
        translate(
            "wizard",
            """The right mouse button also enables one to make a file the
        *main file* of a project. This file can be recognized by its star
        symbol, and it enables running the file more easily.""",
        ),
    ]


class ShellWizardPage1(BasePyzoWizardPage):

    _title = translate("wizard", "The shell is where your code gets executed")
    _image_filename = "pyzo_shell1.png"
    _descriptions = [
        translate(
            "wizard",
            """When Pyzo starts, a default *shell* is created. You can add more
        shells that run simultaneously, and which may be of different
        Python versions.""",
        ),
        translate(
            "wizard",
            """Shells run in a sub-process, such that when it is busy, Pyzo
        itself stays responsive, allowing you to keep coding and even
        run code in another shell.""",
        ),
    ]


class ShellWizardPage2(BasePyzoWizardPage):

    _title = translate("wizard", "Configuring shells")
    _image_filename = "pyzo_shell2.png"
    _descriptions = [
        translate(
            "wizard",
            """Pyzo can integrate the event loop of five different *GUI toolkits*,
        thus enabling interactive plotting with e.g. Visvis or Matplotlib.""",
        ),
        translate(
            "wizard",
            """Via 'Shell > Edit shell configurations', you can edit and add
        *shell configurations*. This allows you to for example select the
        initial directory, or use a custom Pythonpath.""",
        ),
    ]


class RuncodeWizardPage1(BasePyzoWizardPage):

    _title = translate("wizard", "Running code")
    _image_filename = "pyzo_run1.png"
    _descriptions = [
        translate(
            "wizard",
            "Pyzo supports several ways to run source code in the editor. (see the 'Run' menu).",
        ),
        translate(
            "wizard",
            """*Run selection:* if there is no selected text, the current line
        is executed; if the selection is on a single line, the selection
        is evaluated; if the selection spans multiple lines, Pyzo will
        run the the (complete) selected lines.""",
        ),
        translate(
            "wizard",
            "*Run cell:* a cell is everything between two lines starting with '##'.",
        ),
        translate("wizard", "*Run file:* run all the code in the current file."),
        translate(
            "wizard",
            "*Run project main file:* run the code in the current project's main file.",
        ),
    ]


class RuncodeWizardPage2(BasePyzoWizardPage):

    _title = translate("wizard", "Interactive mode vs running as script")
    _image_filename = ""
    _descriptions = [
        translate(
            "wizard",
            """You can run the current file or the main file normally, or as a script.
        When run as script, the shell is restarted to provide a clean
        environment. The shell is also initialized differently so that it
        closely resembles a normal script execution.""",
        ),
        translate(
            "wizard",
            """In interactive mode, sys.path[0] is an empty string (i.e. the current dir),
        and sys.argv is set to [''].""",
        ),
        translate(
            "wizard",
            """In script mode, __file__ and sys.argv[0] are set to the scripts filename,
        sys.path[0] and the working dir are set to the directory containing the script.""",
        ),
    ]


class ToolsWizardPage1(BasePyzoWizardPage):

    _title = translate("wizard", "Tools for your convenience")
    _image_filename = "pyzo_tools1.png"
    _descriptions = [
        translate(
            "wizard",
            """Via the *Tools menu*, one can select which tools to use. The tools can
        be positioned in any way you want, and can also be un-docked.""",
        ),
        translate(
            "wizard",
            """Note that the tools system is designed such that it's easy to
        create your own tools. Look at the online wiki for more information,
        or use one of the existing tools as an example.""",
        ),
    ]


class ToolsWizardPage2(BasePyzoWizardPage):

    _title = translate("wizard", "Recommended tools")
    _image_filename = "pyzo_tools2.png"
    _descriptions = [
        translate("wizard", """We especially recommend the following tools:"""),
        translate(
            "wizard",
            """The *Source structure tool* gives an outline of the source code.""",
        ),
        translate(
            "wizard",
            """The *File browser tool* helps keep an overview of all files
        in a directory. To manage your projects, click the star icon.""",
        ),
    ]


class FinalPage(BasePyzoWizardPage):

    _title = translate("wizard", "Get coding!")
    _image_filename = "pyzologo128.png"
    _descriptions = [
        translate(
            "wizard",
            """This concludes the Pyzo wizard. Now, get coding and have fun!""",
        ),
    ]


# def smooth_images():
#     """ This was used to create the images from their raw versions.
#     """
#
#     import os
#     import visvis as vv
#     import scipy as sp
#     import scipy.ndimage
#     for fname in os.listdir('images'):
#         im = vv.imread(os.path.join('images', fname))
#         for i in range(im.shape[2]):
#             im[:,:,i] = sp.ndimage.gaussian_filter(im[:,:,i], 0.7)
#         #fname = fname.replace('.png', '.jpg')
#         print(fname)
#         vv.imwrite(fname, im[::2,::2,:])


if __name__ == "__main__":
    w = PyzoWizard(None)
    w.show()



# -*- coding: utf-8 -*-
# Copyright (C) 2013, the codeeditor development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

from ..qt import QtCore, QtGui, QtWidgets  # noqa

Qt = QtCore.Qt


class Calltip(object):
    _styleElements = [
        (
            "Editor.calltip",
            "The style of the calltip. ",
            "fore:#555, back:#ff9, border:1",
        )
    ]

    class __CalltipLabel(QtWidgets.QLabel):
        def __init__(self):
            QtWidgets.QLabel.__init__(self)

            # Start hidden
            self.hide()
            # Accept rich text
            self.setTextFormat(QtCore.Qt.RichText)
            # Show as tooltip
            self.setIndent(2)
            self.setWindowFlags(QtCore.Qt.ToolTip)

        def enterEvent(self, event):
            # Act a bit like a tooltip
            self.hide()

    def __init__(self, *args, **kwds):
        super(Calltip, self).__init__(*args, **kwds)
        # Create label for call tips
        self.__calltipLabel = self.__CalltipLabel()
        # Be notified of style updates
        self.styleChanged.connect(self.__afterSetStyle)

        # Prevents calltips from being shown immediately after pressing
        # the escape key.
        self.__noshow = False

    def __afterSetStyle(self):
        format = self.getStyleElementFormat("editor.calltip")
        ss = "QLabel { color:%s; background:%s; border:%ipx solid %s; }" % (
            format["fore"],
            format["back"],
            int(format["border"]),
            format["fore"],
        )
        self.__calltipLabel.setStyleSheet(ss)

    def calltipShow(self, offset=0, richText="", highlightFunctionName=False):
        """calltipShow(offset=0, richText='', highlightFunctionName=False)

        Shows the given calltip.

        Parameters
        ----------
        offset : int
            The character offset to show the tooltip.
        richText : str
            The text to show (may contain basic html for markup).
        highlightFunctionName : bool
            If True the text before the first opening brace is made bold.
            default False.

        """

        # Do not show the calltip if it was deliberately hidden by the
        # user.
        if self.__noshow:
            return

        # Process calltip text?
        if highlightFunctionName:
            i = richText.find("(")
            if i > 0:
                richText = "<b>{}</b>{}".format(richText[:i], richText[i:])

        # Get a cursor to establish the position to show the calltip
        startcursor = self.textCursor()
        startcursor.movePosition(startcursor.Left, n=offset)

        # Get position in pixel coordinates
        rect = self.cursorRect(startcursor)
        pos = rect.topLeft()
        pos.setY(pos.y() - rect.height() - 1)  # Move one above line
        pos.setX(pos.x() - 3)  # Correct for border and indent
        pos = self.viewport().mapToGlobal(pos)

        # Set text and update font
        self.__calltipLabel.setText(richText)
        self.__calltipLabel.setFont(self.font())

        # Use a qt tooltip to show the calltip
        if richText:
            self.__calltipLabel.move(pos)
            self.__calltipLabel.show()
        else:
            self.__calltipLabel.hide()

    def calltipCancel(self):
        """calltipCancel()

        Hides the calltip.

        """
        self.__calltipLabel.hide()

    def calltipActive(self):
        """calltipActive()

        Get whether the calltip is currently active.

        """
        return self.__calltipLabel.isVisible()

    def focusOutEvent(self, event):
        super(Calltip, self).focusOutEvent(event)
        self.__calltipLabel.hide()

    def keyPressEvent(self, event):
        # If the user presses Escape and the calltip is active, hide it
        if (
            event.key() == Qt.Key_Escape
            and event.modifiers() == Qt.NoModifier
            and self.calltipActive()
        ):
            self.calltipCancel()
            self.__noshow = True
            return

        if event.text() == "(":
            self.__noshow = False
        elif event.text() == ")":
            self.calltipCancel()

        # Proceed processing the keystrike
        super(Calltip, self).keyPressEvent(event)



"""
Small script to check the log generated by a test run (either frozen or source).
"""

import os
import sys

# Locate the log
logfilename = os.path.abspath(os.path.join(__file__, "..", "..", "log.txt"))
if not os.path.isfile(logfilename):
    raise RuntimeError(f"Pyzo log file not found in {logfilename}")

# Read and clear the log
with open(logfilename, "rt") as f:
    log = f.read()
os.remove(logfilename)

# Print the log
print("=" * 80)
print(log)

# Examine the results
if any(x in log.lower() for x in ["exception", "uncaught", "error"]):
    sys.exit("Errors detected during Pyzo test run :(")
elif not log.strip().endswith("Stopped"):
    sys.exit("Unclean stop for Pyzo test run :(")
else:
    print("==> Pyzo test run looks OK :)")
    sys.exit(0)



# -*- coding: utf-8 -*-
# Copyright (C) 2016, the Pyzo development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

""" Module kernelBroker

This module implements the interface between Pyzo and the kernel.

"""

import os, sys, time
import subprocess
import signal
import threading
import ctypes

import yoton
import pyzo  # local Pyzo (can be on a different box than where the user is)
from pyzo.util import zon as ssdf  # zon is ssdf-light


# To allow interpreters relative to (frozen) Pyzo app
EXE_DIR = os.path.abspath(os.path.dirname(sys.executable))
if EXE_DIR.endswith(".app/Contents/MacOS"):
    EXE_DIR = os.path.dirname(EXE_DIR.rsplit(".app")[0])


# Important: the yoton event loop should run somehow!


class KernelInfo(ssdf.Struct):
    """KernelInfo

    Describes all information for a kernel. This class can be used at
    the IDE as well as the kernelbroker.

    This information goes a long way from the pyzo config file to the
    kernel. The list pyzo.config.shellConfigs2 contains the configs
    for all kernels. These objects are edited in-place by the
    shell config.

    The shell keeps a reference of the shell config used to start the
    kernel. On each restart all information is resend. In this way,
    if a user changes a setting in the shell config, it is updated
    when the shell restarts.

    The broker also keeps a copy of the shell config. In this way,
    the shell might send no config information (or only partially
    update the config information) on a restart. This is not so
    relevant now, but it can be when we are running multiple people
    on a single kernel, and there is only one user who has the
    original config.

    """

    def __init__(self, info=None):
        super().__init__()
        # ----- Fixed parameters that define a shell -----

        # scriptFile is used to define the mode. If given, we run in
        # script-mode. Otherwise we run in interactive mode.

        # The name of this shell config. Can be used to name the kernel
        self.name = "Python"

        # The executable. This can be '/usr/bin/python3.1' or
        # 'c:/program files/python2.6/python.exe', etc.
        self.exe = ""

        # The GUI toolkit to embed the event loop of.
        # Instantiate with a value that is settable
        self.gui = "Auto"

        # The Python path. Paths should be separated by newlines.
        # '$PYTHONPATH' is replaced by environment variable by broker
        self.pythonPath = ""

        # The path of the current project, the kernel will prepend this
        # to the sys.path. The broker could prepend to PYTHONPATH, but
        # in this way it is more explicit (kernel can tell the user that
        # the project path was prepended).
        self.projectPath = ""

        # The full filename of the script to run.
        # If given, the kernel should run in script-mode.
        # The kernel will check whether this file exists, and will
        # revert to interactive mode if it doesn't.
        self.scriptFile = ""

        # Interactive-mode only:

        # The initial directory. Only used for interactive-mode; in
        # script-mode the initial directory is the dir of the script.
        self.startDir = ""

        # The Startup script (only used for interactive-mode).
        # - Empty string means run nothing,
        # - Single line means file name
        # - multiple lines means source code.
        # - '$PYTHONSTARTUP' uses the code in that file. Broker replaces this.
        self.startupScript = ""

        # Additional command line arguments, set by the kernel
        self.argv = ""

        # Additional environment variables
        self.environ = ""

        # Load info from ssdf struct. Make sure they are all strings
        if info:
            # Get struct
            if isinstance(info, dict):
                s = info
            elif isinstance(info, str):
                s = ssdf.loads(info)
            else:
                raise ValueError(
                    "Kernel info should be a string or ssdf struct, not %s"
                    % str(type(info))
                )
            # Inject values
            for key in s:
                val = s[key]
                if not val:
                    val = ""
                self[key] = val

    def tostring(self):
        return ssdf.saves(self)


def getCommandFromKernelInfo(info, port):
    info = KernelInfo(info)

    # Apply default exe
    exe = info.exe or "python"
    if exe.startswith("."):
        exe = os.path.abspath(os.path.join(EXE_DIR, exe))

    # Correct path when it contains spaces
    if exe.count(" ") and exe[0] != '"':
        exe = '"{}"'.format(exe)

    # Get start script
    startScript = os.path.join(pyzo.pyzoDir, "pyzokernel", "start.py")
    startScript = '"{}"'.format(startScript)

    # Build command
    command = exe + " " + startScript + " " + str(port)

    # Done
    return command


def getEnvFromKernelInfo(info):
    info = KernelInfo(info)

    pythonPath = info.pythonPath

    # Set default pythonPath (replace only first occurrence of $PYTHONPATH
    ENV_PP = os.environ.get("PYTHONPATH", "")
    pythonPath = pythonPath.replace("$PYTHONPATH", "\n" + ENV_PP + "\n", 1)
    pythonPath = pythonPath.replace("$PYTHONPATH", "")
    # Split paths, allow newlines and os.pathsep
    for splitChar in "\n\r" + os.pathsep:
        pythonPath = pythonPath.replace(splitChar, "\n")
    pythonPaths = [p.strip() for p in pythonPath.split("\n") if p]
    # Recombine using the OS's path separator
    pythonPath = os.pathsep.join(pythonPaths)
    # Add entry to Pythopath, so that we can import yoton
    # Note: an empty entry might cause trouble if the start-directory is
    # somehow overriden (see issue 128).
    pythonPath = pyzo.pyzoDir + os.pathsep + pythonPath

    # Prepare environment, remove references to tk libraries,
    # since they're wrong when frozen. Python will insert the
    # correct ones if required.
    env = os.environ.copy()
    #
    env.pop("TK_LIBRARY", "")
    env.pop("TCL_LIBRARY", "")
    env["PYTHONPATH"] = pythonPath
    # Jython does not use PYTHONPATH but JYTHONPATH
    env["JYTHONPATH"] = pyzo.pyzoDir + os.pathsep + os.environ.get("JYTHONPATH", "")
    env["TERM"] = "dumb"  # we have a "dumb" terminal (see #422)

    # PyInstaller uses DllDirectory, but this leaks down to subprocesses,
    # causing GUI integration with QT-based libs to fail. Turn it off now.
    # https://github.com/pyinstaller/pyinstaller/issues/3795
    if sys.platform.startswith("win32"):
        ctypes.windll.kernel32.SetDllDirectoryA(None)

    # PyInstaller prepends the root app dir to LD_LIBRARY_PATH (see #665)
    if getattr(sys, "frozen", False) and "LD_LIBRARY_PATH" in env:
        to_remove = os.path.normpath(os.path.dirname(sys.executable))
        paths = env["LD_LIBRARY_PATH"].split(os.pathsep)
        paths = [os.path.normpath(p) for p in paths]
        paths = [p for p in paths if p != to_remove]
        env["LD_LIBRARY_PATH"] = os.pathsep.join(paths)

    # Remove Qt plugin directories, because it breaks Qt integration for the kernel
    # on several systems. E.g. QT_QPA_PLATFORM_PLUGIN_PATH, QT_PLUGIN_PATH
    if getattr(sys, "frozen", False):
        frozen_path = os.path.normpath(sys.prefix).lower()
        for key, val in list(env.items()):
            if key.startswith(("QT_", "QML2_")):
                if frozen_path in os.path.normpath(val).lower():
                    env.pop(key, None)
        env["PATH"] = os.pathsep.join(
            x
            for x in os.getenv("PATH", "").split(os.pathsep)
            if not os.path.normpath(x).lower().startswith(frozen_path)
        )

    # Add dirs specific to this Python interpreter. Especially important with
    # Miniconda/Anaconda on Windows, see issue #591
    prefix = os.path.dirname(info.exe)
    envdirs = ["", "Library/usr/bin", "Library/bin", "bin"]
    if sys.platform.startswith("win"):
        envdirs.extend([r"Scripts", r"Library\mingw-w64\bin", r"Library\mingw-w32\bin"])
    curpath = env.get("PATH", "").strip(os.pathsep)
    env["PATH"] = (
        os.pathsep.join(os.path.join(prefix, d) for d in envdirs) + os.pathsep + curpath
    )

    # Add environment variables specified in shell config
    for line in info.environ.splitlines():
        line = line.strip()
        if "=" in line:
            key, val = line.split("=", 1)
            if key:
                key, val = key.strip(), val.strip()
                env[key] = os.path.expandvars(val)

    # Done
    return env


class KernelBroker:
    """KernelBroker(info)

    This class functions as a broker between a kernel process and zero or
    more IDE's (clients).

    This class has a single context assosiated with it, that lives as long
    as this object. It is used to connect to a kernel process and to
    0 or more IDE's (clients). The kernel process can be "restarted", meaning
    that it is terminated and a new process started.

    The broker is cleaned up if there is no kernel process AND no connections.

    """

    def __init__(self, manager, info, name=""):
        self._manager = manager

        # Store info that defines the kernel
        self._originalInfo = KernelInfo(info)

        # Make a copy for the current version. This copy is re-created on
        # each restart
        self._info = ssdf.copy(self._originalInfo)

        # Store name (or should the name be defined in the info struct)
        self._name = name

        # Create context for the connection to the kernel and IDE's
        # This context is persistent (it stays as long as this KernelBroker
        # instance is alive).
        self._context = yoton.Context()
        self._kernelCon = None
        self._ctrl_broker = None

        # Create yoton-based timer
        self._timer = yoton.Timer(0.2, oneshot=False)
        self._timer.bind(self.mainLoopIter)

        # Kernel process and connection (these are replaced on restarting)
        self._reset()

        # For restarting after terminating
        self._pending_restart = None

    ## Startup and teardown

    def _create_channels(self):
        ct = self._context

        # Close any existing channels first
        self._context.close_channels()

        # Create stream channels.
        # Stdout is for the C-level stdout/stderr streams.
        self._strm_broker = yoton.PubChannel(ct, "strm-broker")
        self._strm_raw = yoton.PubChannel(ct, "strm-raw")
        self._strm_prompt = yoton.PubChannel(ct, "strm-prompt")

        # Create control channel so that the IDE can control restarting etc.
        self._ctrl_broker = yoton.SubChannel(ct, "ctrl-broker")

        # Status channel to pass startup parameters to the kernel
        self._stat_startup = yoton.StateChannel(ct, "stat-startup", yoton.OBJECT)

        # We use the stat-interpreter to set the status to dead when kernel dies
        self._stat_interpreter = yoton.StateChannel(ct, "stat-interpreter")

        # Create introspect channel so we can interrupt and terminate
        self._reqp_introspect = yoton.ReqChannel(ct, "reqp-introspect")

    def _reset(self, destroy=False):
        """_reset(destroy=False)

        Reset state. if destroy, does a full clean up, closing the context
        and removing itself from the KernelManager's list.

        """

        # Close connection (it might be in a wait state if the process
        # failed to start)
        if self._kernelCon is not None:
            self._kernelCon.close()

        # Set process and kernel connection to None
        self._process = None
        self._kernelCon = None
        self._terminator = None
        self._streamReader = None

        if destroy:

            # Stop timer
            self._timer.unbind(self.mainLoopIter)
            self._timer.stop()
            self._timer = None

            # Clean up this kernelbroker instance
            L = self._manager._kernels
            while self in L:
                L.remove(self)

            # Remove references
            #
            if self._context is not None:
                self._context.close()
            self._context = None
            #
            self._strm_broker = None
            self._strm_raw = None
            self._stat_startup = None
            self._stat_interpreter = None
            self._strm_prompt = None
            #
            self._ctrl_broker = None
            self._reqp_introspect = None

    def startKernelIfConnected(self, timeout=10.0):
        """startKernelIfConnected(timout=10.0)

        Start the kernel as soon as there is a connection.

        """
        self._process = time.time() + timeout
        self._timer.start()

    def startKernel(self):
        """startKernel()

        Launch the kernel in a subprocess, and connect to it via the
        context and two Pypes.

        """

        # Create channels
        self._create_channels()

        # Create info dict
        info = {}
        for key in self._info:
            info[key] = self._info[key]

        # Send info stuff so that the kernel has access to the information
        self._stat_startup.send(info)

        # Get directory to start process in
        cwd = pyzo.pyzoDir

        # Host connection for the kernel to connect
        # (tries several port numbers, staring from 'PYZO')
        self._kernelCon = self._context.bind(
            "localhost:PYZO", max_tries=256, name="kernel"
        )

        # Get command to execute, and environment to use
        command = getCommandFromKernelInfo(self._info, self._kernelCon.port1)
        env = getEnvFromKernelInfo(self._info)

        # Wrap command in call to 'cmd'?
        if sys.platform.startswith("win"):
            # as the author from Pype writes:
            # if we don't run via a command shell, then either sometimes we
            # don't get wx GUIs, or sometimes we can't kill the subprocesses.
            # And I also see problems with Tk.
            # But we only use it if we are sure that cmd is available.
            # See pyzo issue #240
            try:
                subprocess.check_output('cmd /c "cd"', shell=True)
            except (IOError, subprocess.SubprocessError):
                pass  # Do not use cmd
            else:
                command = 'cmd /c "{}"'.format(command)

        # Start process
        self._process = subprocess.Popen(
            command,
            shell=True,
            env=env,
            cwd=cwd,
            stdin=subprocess.PIPE,  # Fixes issue 165
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        # Set timeout for connection, i.e. after how much time of
        # unresponsive ness is the kernel found to be running extension code
        # Better set this before connecting
        self._kernelCon.timeout = 0.5

        # Bind to events
        self._kernelCon.closed.bind(self._onKernelConnectionClose)
        self._kernelCon.timedout.bind(self._onKernelTimedOut)

        # Create reader for stream
        self._streamReader = StreamReader(
            self._process, self._strm_raw, self._strm_broker
        )

        # Start streamreader and timer
        self._streamReader.start()
        self._timer.start()

        # Reset some variables
        self._pending_restart = None

    def hostConnectionForIDE(self, address="localhost"):
        """hostConnectionForIDE()

        Host a connection for an IDE to connect to. Returns the port to which
        the ide can connect.

        """
        c = self._context.bind(address + ":PYZO+256", max_tries=256, name="ide")
        return c.port1

    ## Callbacks

    def _onKernelTimedOut(self, c, timedout):
        """_onKernelTimedOut(c, timeout)

        The kernel timed out (i.e. did not send heartbeat messages for
        a while. It is probably running extension code.

        """
        if timedout:
            self._stat_interpreter.send("Very busy")
        else:
            self._stat_interpreter.send("Busy")

    def _onKernelConnectionClose(self, c, why):
        """_onKernelConnectionClose(c, why)

        Connection with kernel lost. Tell clients why.

        """

        # If we receive this event while the current kernel connection
        # is not the one that generated the event, ignore it.
        if self._kernelCon is not c:
            return

        # The only reasonable way that the connection
        # can be lost without the kernel closing, is if the yoton context
        # crashed or was stopped somehow. In both cases, we lost control,
        # and should put it down!
        if not self._terminator:
            self.terminate("because connecton was lost", "KILL", 0.5)

    def _onKernelDied(self, returncode=0):
        """_onKernelDied()

        Kernel process died. Clean up!

        """

        # If the kernel did not start yet, probably the command is invalid
        if self._kernelCon and self._kernelCon.is_waiting:
            msg = "The process failed to start (invalid command?)."
        elif not self.isTerminating():
            msg = "Kernel process exited."
        elif not self._terminator._prev_action:
            # We did not actually take any terminating action
            # This happens, because if the kernel is killed from outside,
            # _onKernelConnectionClose() triggers a terminate sequence
            # (but with a delay).
            # Note the "The" to be able to distinguish this case
            msg = "The kernel process exited."
        else:
            msg = self._terminator.getMessage("Kernel process")

        if self._context.connection_count:
            # Notify
            returncodeMsg = "\n%s (%s)\n\n" % (msg, str(returncode))
            self._strm_broker.send(returncodeMsg)
            # Empty prompt and signal dead
            self._strm_prompt.send("\b")
            self._stat_interpreter.send("Dead")
            self._context.flush()

        # Cleanup (get rid of kernel process references)
        self._reset()

        # Handle any pending action
        if self._pending_restart:
            self.startKernel()

    ## Main loop and termination

    def terminate(self, reason="by user", action="TERM", timeout=0.0):
        """terminate(reason='by user', action='TERM', timeout=0.0)

        Initiate termination procedure for the current kernel.

        """

        # The terminatation procedure is started by creating
        # a KernelTerminator instance. This instance's iteration method
        # iscalled from _mailLoopIter().
        self._terminator = KernelTerminator(self, reason, action, timeout)

    def isTerminating(self):
        """isTerminating()

        Get whether the termination procedure has been initiated. This
        simply checks whether there is a self._terminator instance.

        """
        return bool(self._terminator)

    def mainLoopIter(self):
        """mainLoopIter()

        Periodically called. Kind of the main loop iteration for this kernel.

        """

        # Get some important status info
        hasProcess = self._process is not None
        hasKernelConnection = bool(self._kernelCon and self._kernelCon.is_connected)
        hasClients = False
        if self._context:
            hasClients = self._context.connection_count > int(hasKernelConnection)

        # Should we clean the whole thing up?
        if not (hasProcess or hasClients):
            self._reset(True)  # Also unregisters this timer callback
            return

        # Waiting to get started; waiting for client to connect
        if isinstance(self._process, float):
            if self._context.connection_count:
                self.startKernel()
            elif self._process > time.time():
                self._process = None
            return

        # If we have a process ...
        if self._process:
            # Test if process is dead
            process_returncode = self._process.poll()
            if process_returncode is not None:
                self._onKernelDied(process_returncode)
                return
            # Are we in the process of terminating?
            elif self.isTerminating():
                self._terminator.next()
        elif self.isTerminating():
            # We cannot have a terminator if we have no process
            self._terminator = None

        # handle control messages
        if self._ctrl_broker:
            for msg in self._ctrl_broker.recv_all():
                if msg == "INT":
                    self._commandInterrupt()
                elif msg == "TERM":
                    self._commandTerminate()
                elif msg.startswith("RESTART"):
                    self._commandRestart(msg)
                else:
                    pass  # Message is not for us

    def _commandInterrupt(self):
        if self._process is None:
            self._strm_broker.send("Cannot interrupt: process is dead.\n")
        # Kernel receives and acts
        elif sys.platform.startswith("win"):
            self._reqp_introspect.interrupt()
        else:
            # Use POSIX to interrupt, which is more reliable
            # (the introspect thread might not get a chance)
            # but also does not work if running extension code
            pid = self._kernelCon.pid2
            os.kill(pid, signal.SIGINT)

    def _commandTerminate(self):
        # Start termination procedure
        # Kernel will receive term and act (if it can).
        # If it wont, we will act in a second or so.
        if self._process is None:
            self._strm_broker.send("Cannot terminate: process is dead.\n")
        elif self.isTerminating():
            # The user gave kill command while the kill process
            # is running. We could do an immediate kill now,
            # or we let the terminate process run its course.
            pass
        else:
            self.terminate("by user")

    def _commandRestart(self, msg):
        # Almost the same as terminate, but now we have a pending action
        self._pending_restart = True

        # Recreate the info struct
        self._info = ssdf.copy(self._originalInfo)
        # Update the info struct
        new_info = ssdf.loads(msg.split("RESTART", 1)[1])
        for key in new_info:
            self._info[key] = new_info[key]

        # Restart now, wait, or initiate termination procedure?
        if self._process is None:
            self.startKernel()
        elif self.isTerminating():
            pass  # Already terminating
        else:
            self.terminate("for restart")


class KernelTerminator:
    """KernelTerminator(broker, reason='user terminated', action='TERM', timeout=0.0)

    Simple class to help terminating the kernel. It has a next() method
    that should be periodically called. It keeps track whether the timeout
    has passed and will undertake increaslingly ruder actions to terminate
    the kernel.

    """

    def __init__(self, broker, reason="by user", action="TERM", timeout=0.0):

        # Init/store
        self._broker = broker
        self._reason = reason
        self._next_action = ""

        # Go
        self._do(action, timeout)

    def _do(self, action, timeout):
        self._prev_action = self._next_action
        self._next_action = action
        self._timeout = time.time() + timeout
        if not timeout:
            self.next()

    def next(self):

        # Get action
        action = self._next_action

        if time.time() < self._timeout:
            # Time did not pass yet
            pass

        elif action == "TERM":
            self._broker._reqp_introspect.terminate()
            self._do("INT", 0.5)

        elif action == "INT":
            # Count
            if not hasattr(self, "_count"):
                self._count = 0
            self._count += 1
            # Handle
            if self._count < 5:
                self._broker._reqp_introspect.interrupt()
                self._do("INT", 0.1)
            else:
                self._do("KILL", 0)

        elif action == "KILL":
            # Get pid and signal
            pid = self._broker._kernelCon.pid2
            sigkill = signal.SIGTERM
            if hasattr(signal, "SIGKILL"):
                sigkill = signal.SIGKILL
            # Kill
            if hasattr(os, "kill"):
                os.kill(pid, sigkill)
            elif sys.platform.startswith("win"):
                kernel32 = ctypes.windll.kernel32
                handle = kernel32.OpenProcess(1, 0, pid)
                kernel32.TerminateProcess(handle, 0)
                # os.system("TASKKILL /PID " + str(pid) + " /F")
            # Set what we did
            self._do("NOTHING", 9999999999999999)

    def getMessage(self, what):
        # Get nice string of that
        D = {
            "": "exited",
            "TERM": "terminated",
            "INT": "terminated (after interrupting)",
            "KILL": "killed",
        }
        actionMsg = D.get(self._prev_action, "stopped for unknown reason")

        # Compile stop-string
        return "{} {} {}.".format(what, actionMsg, self._reason)


class StreamReader(threading.Thread):
    """StreamReader(process, channel)

    Reads stdout of process and send to a yoton channel.
    This needs to be done in a separate thread because reading from
    a PYPE blocks.

    """

    def __init__(self, process, strm_raw, strm_broker):
        threading.Thread.__init__(self)

        self._process = process
        self._strm_raw = strm_raw
        self._strm_broker = strm_broker
        self.deamon = True
        self._exit = False

    def stop(self, timeout=1.0):
        self._exit = True
        self.join(timeout)

    def run(self):
        while not self._exit:
            time.sleep(0.001)
            # Read any stdout/stderr messages and route them via yoton.
            msg = self._process.stdout.readline()  # <-- Blocks here
            if not isinstance(msg, str):
                msg = msg.decode("utf-8", "ignore")
            try:
                self._strm_raw.send(msg)
            except IOError:
                pass  # Channel is closed
            # Process dead?
            if not msg:  # or self._process.poll() is not None:
                break
        # self._strm_broker.send('streamreader exit\n')


class Kernelmanager:
    """Kernelmanager

    This class manages a set of kernels. These kernels run on the
    same machine as this broker. IDE's can ask which kernels are available
    and can connect to them via this broker.

    The Pyzo process runs an instance of this class that connects at
    localhost. At a later stage, we may make it possible to create
    a kernel-server at a remote machine.

    """

    def __init__(self, public=False):

        # Set whether other machines in this network may connect to our kernels
        self._public = public

        # Init list of kernels
        self._kernels = []

    def createKernel(self, info, name=None):
        """create_kernel(info, name=None)

        Create a new kernel. Returns the port number to connect to the
        broker's context.

        """

        # Set name if not given
        if not name:
            i = len(self._kernels) + 1
            name = "kernel %i" % i

        # Create kernel
        kernel = KernelBroker(self, info, name)
        self._kernels.append(kernel)

        # Host a connection for the ide
        port = kernel.hostConnectionForIDE()

        # Tell broker to start as soon as the IDE connects with the broker
        kernel.startKernelIfConnected()

        # Done
        return port

    def getKernelList(self):

        # Get info of each kernel as an ssdf struct
        infos = []
        for kernel in self._kernels:
            info = kernel._info
            info = ssdf.loads(info.tostring())
            info.name = kernel._name
            infos.append(info)

        # Done
        return infos

    def terminateAll(self):
        """terminateAll()

        Terminates all kernels. Required when shutting down Pyzo.
        When this function returns, all kernels will be terminated.

        """
        for kernel in [kernel for kernel in self._kernels]:

            # Try closing the process gently: by closing stdin
            terminator = KernelTerminator(kernel, "for closing down")

            # Terminate
            while (
                kernel._kernelCon
                and kernel._kernelCon.is_connected
                and kernel._process
                and (kernel._process.poll() is None)
            ):
                terminator.next()
                time.sleep(0.02)
                yoton.process_events(False)

            # Clean up
            kernel._reset(True)



# -*- coding: utf-8 -*-
# Copyright (C) 2016, the Pyzo development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.


""" Module shellStack

Implements the stack of shells. Also implements the nifty debug button
and a dialog to edit the shell configurations.

"""

import time
import webbrowser
from pyzo.qt import QtCore, QtGui, QtWidgets  # noqa

import pyzo
from pyzo import translate
from pyzo.core.shell import PythonShell
from pyzo.core.pyzoLogging import print  # noqa
from pyzo.core.menu import ShellTabContextMenu, ShellButtonMenu
from pyzo.core.icons import ShellIconMaker


def shellTitle(shell, moreinfo=False):
    """Given a shell instance, build the text title to represent it."""

    # Get name
    nameText = shell._info.name

    # Build version text
    if shell._version:
        versionText = "v{}".format(shell._version)
    else:
        versionText = "v?"

    # Build gui text
    guiText = shell._startup_info.get("gui")
    guiText = guiText or ""
    if guiText.lower() in ["none", ""]:
        guiText = "without gui"
    else:
        guiText = "with " + guiText + " gui"

    # Build state text
    stateText = shell._state or ""

    # Build text for elapsed time
    elapsed = time.time() - shell._start_time
    hh = elapsed // 3600
    mm = (elapsed - hh * 3600) // 60
    ss = elapsed - hh * 3600 - mm * 60
    runtimeText = "runtime: %i:%02i:%02i" % (hh, mm, ss)

    # Build text
    if not moreinfo:
        text = nameText
    else:
        text = "'%s' (%s %s) - %s, %s" % (
            nameText,
            versionText,
            guiText,
            stateText,
            runtimeText,
        )

    # Done
    return text


class ShellStackWidget(QtWidgets.QWidget):
    """The shell stack widget provides a stack of shells.

    It wrapps a QStackedWidget that contains the shell objects. This
    stack is used as a reference to synchronize the shell selection with.
    We keep track of what is the current selected shell and apply updates
    if necessary. Therefore, changing the current shell in the stack
    should be enough to invoke a full update.

    """

    # When the current shell changes.
    currentShellChanged = QtCore.Signal()

    # When the current shells state (or debug state) changes,
    # or when a new prompt is received.
    # Also fired when the current shell changes.
    currentShellStateChanged = QtCore.Signal()

    def __init__(self, parent):
        QtWidgets.QWidget.__init__(self, parent)

        # create toolbar
        self._toolbar = QtWidgets.QToolBar(self)
        self._toolbar.setMaximumHeight(26)
        self._toolbar.setIconSize(QtCore.QSize(16, 16))

        # create stack
        self._stack = QtWidgets.QStackedWidget(self)

        # Populate toolbar
        self._shellButton = ShellControl(self._toolbar, self._stack)
        self._debugmode = 0
        self._dbs = DebugStack(self._toolbar)
        #
        self._toolbar.addWidget(self._shellButton)
        self._toolbar.addSeparator()
        # self._toolbar.addWidget(self._dbc) -> delayed, see addContextMenu()

        self._interpreterhelp = InterpreterHelper(self)

        # widget layout
        layout = QtWidgets.QVBoxLayout()
        layout.setSpacing(0)
        # set margins
        margin = pyzo.config.view.widgetMargin
        layout.setContentsMargins(margin, margin, margin, margin)

        layout.addWidget(self._toolbar)
        layout.addWidget(self._stack, 0)
        layout.addWidget(self._interpreterhelp, 0)
        self.setLayout(layout)

        # make callbacks
        self._stack.currentChanged.connect(self.onCurrentChanged)

        self.showInterpreterHelper()

    def __iter__(self):
        i = 0
        while i < self._stack.count():
            w = self._stack.widget(i)
            i += 1
            yield w

    def showInterpreterHelper(self, show=True):
        self._interpreterhelp.setVisible(show)
        self._toolbar.setVisible(not show)
        self._stack.setVisible(not show)
        if show:
            self._interpreterhelp.detect()

    def addShell(self, shellInfo=None):
        """addShell()
        Add a shell to the widget."""

        # Create shell and add to stack
        shell = PythonShell(self, shellInfo)
        self._stack.addWidget(shell)
        # Bind to signals
        shell.stateChanged.connect(self.onShellStateChange)
        shell.debugStateChanged.connect(self.onShellDebugStateChange)
        # Select it and focus on it (invokes onCurrentChanged)
        self._stack.setCurrentWidget(shell)
        shell.setFocus()
        return shell

    def removeShell(self, shell):
        """removeShell()
        Remove an existing shell from the widget
        """
        self._stack.removeWidget(shell)

    def onCurrentChanged(self, index):
        """When another shell is selected, update some things."""

        # Get current
        shell = self.getCurrentShell()
        # Call functions
        self.onShellStateChange(shell)
        self.onShellDebugStateChange(shell)
        # Emit Signal
        self.currentShellChanged.emit()

    def onShellStateChange(self, shell):
        """Called when the shell state changes, and is called
        by onCurrentChanged. Sets the mainwindow's icon if busy.
        """

        # Keep shell button and its menu up-to-date
        self._shellButton.updateShellMenu(shell)

        if shell is self.getCurrentShell():  # can be None
            # Update application icon
            if shell and shell._state in ["Busy"]:
                pyzo.main.setWindowIcon(pyzo.iconRunning)
            else:
                pyzo.main.setWindowIcon(pyzo.icon)
            # Send signal
            self.currentShellStateChanged.emit()

    def onShellDebugStateChange(self, shell):
        """Called when the shell debug state changes, and is called
        by onCurrentChanged. Sets the debug button.
        """

        if shell is self.getCurrentShell():

            # Update debug info
            if shell and shell._debugState:
                info = shell._debugState
                self._debugmode = info["debugmode"]
                for action in self._debugActions:
                    action.setEnabled(self._debugmode == 2)
                self._debugActions[-1].setEnabled(self._debugmode > 0)  # Stop
                self._dbs.setTrace(shell._debugState)
            else:
                for action in self._debugActions:
                    action.setEnabled(False)
                self._debugmode = 0
                self._dbs.setTrace(None)
            # Send signal
            self.currentShellStateChanged.emit()

    def getCurrentShell(self):
        """getCurrentShell()
        Get the currently active shell.
        """

        w = None
        if self._stack.count():
            w = self._stack.currentWidget()
        if not w:
            return None
        else:
            return w

    def getShells(self):
        """Get all shell in stack as list"""

        shells = []
        for i in range(self._stack.count()):
            shell = self.getShellAt(i)
            if shell is not None:
                shells.append(shell)

        return shells

    def getShellAt(self, i):
        return
        """ Get shell at current tab index """

        return self._stack.widget(i)

    def addContextMenu(self):
        # A bit awkward... but the ShellMenu needs the ShellStack, so it
        # can only be initialized *after* the shellstack is created ...

        # Give shell tool button a menu
        self._shellButton.setMenu(ShellButtonMenu(self, "Shell button menu"))
        self._shellButton.menu().aboutToShow.connect(
            self._shellButton._elapsedTimesTimer.start
        )

        # Also give it a context menu
        self._shellButton.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self._shellButton.customContextMenuRequested.connect(self.contextMenuTriggered)

        # Add actions
        for action in pyzo.main.menuBar()._menumap["shell"]._shellActions:
            action = self._toolbar.addAction(action)

        self._toolbar.addSeparator()

        # Add debug actions
        self._debugActions = []
        for action in pyzo.main.menuBar()._menumap["shell"]._shellDebugActions:
            self._debugActions.append(action)
            action = self._toolbar.addAction(action)

        # Delayed-add debug control buttons
        self._toolbar.addWidget(self._dbs)

    def contextMenuTriggered(self, p):
        """Called when context menu is clicked"""

        # Get index of shell belonging to the tab
        shell = self.getCurrentShell()

        if shell:
            p = self._shellButton.mapToGlobal(self._shellButton.rect().bottomLeft())
            ShellTabContextMenu(shell=shell, parent=self).popup(p)

    def onShellAction(self, action):
        shell = self.getCurrentShell()
        if shell:
            getattr(shell, action)()


class ShellControl(QtWidgets.QToolButton):
    """A button that can be used to select a shell and start a new shell."""

    def __init__(self, parent, shellStack):
        QtWidgets.QToolButton.__init__(self, parent)

        # Store reference of shell stack
        self._shellStack = shellStack

        # Keep reference of actions corresponding to shells
        self._shellActions = []

        # Set text and tooltip
        self.setText("Warming up ...")
        self.setToolTip(translate("shells", "Click to select shell."))
        self.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        self.setPopupMode(self.InstantPopup)

        # Set icon
        self._iconMaker = ShellIconMaker(self)
        self._iconMaker.updateIcon("busy")  # Busy initializing

        # Create timer
        self._elapsedTimesTimer = QtCore.QTimer(self)
        self._elapsedTimesTimer.setInterval(1000)
        self._elapsedTimesTimer.setSingleShot(False)
        self._elapsedTimesTimer.timeout.connect(self.onElapsedTimesTimer)

    def updateShellMenu(self, shellToUpdate=None):
        """Update the shell menu. Ensure that there is a menu item
        for each shell. If shellToUpdate is given, updates the corresponding
        menu item.
        """
        menu = self.menu()

        # Get shells now active
        currentShell = self._shellStack.currentWidget()
        shells = [self._shellStack.widget(i) for i in range(self._shellStack.count())]

        # Synchronize actions. Remove invalid actions
        for action in self._shellActions:
            # Check match with shells
            if action._shell in shells:
                shells.remove(action._shell)
            else:
                menu.removeAction(action)
            # Update checked state
            if action._shell is currentShell and currentShell:
                action.setChecked(True)
            else:
                action.setChecked(False)
            # Update text if necessary
            if action._shell is shellToUpdate:
                action.setText(shellTitle(shellToUpdate, True))

        # Any items left in shells need a menu item
        # Dont give them an icon, or the icon is used as checkbox thingy
        for shell in shells:
            text = shellTitle(shell)
            action = menu.addItem(text, None, self._shellStack.setCurrentWidget, shell)
            action._shell = shell
            action.setCheckable(True)
            self._shellActions.append(action)

        # Is the shell being updated the current?
        if currentShell is shellToUpdate and currentShell is not None:
            self._iconMaker.updateIcon(currentShell._state)
            self.setText(shellTitle(currentShell))
        elif currentShell is None:
            self._iconMaker.updateIcon("")
            self.setText("No shell selected")

    def onElapsedTimesTimer(self):
        # Automatically turn timer off is menu is hidden
        if not self.menu().isVisible():
            self._elapsedTimesTimer.stop()
            return

        # Update text for each shell action
        for action in self._shellActions:
            action.setText(shellTitle(action._shell, True))


# todo: remove this?
# class DebugControl(QtWidgets.QToolButton):
#     """ A button to control debugging.
#     """
#
#     def __init__(self, parent):
#         QtWidgets.QToolButton.__init__(self, parent)
#
#         # Flag
#         self._debugmode = False
#
#         # Set text
#         self.setText(translate('debug', 'Debug'))
#         self.setIcon(pyzo.icons.bug)
#         self.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
#         #self.setPopupMode(self.InstantPopup)
#
#         # Bind to triggers
#         self.triggered.connect(self.onTriggered)
#         self.pressed.connect(self.onPressed)
#         self.buildMenu()
#
#
#     def buildMenu(self):
#
#         # Count breakpoints
#         bpcount = 0
#         for e in pyzo.editors:
#             bpcount += len(e.breakPoints())
#
#         # Prepare a text
#         clearallbps = translate('debug', 'Clear all {} breakpoints')
#         clearallbps = clearallbps.format(bpcount)
#
#         # Set menu
#         menu = QtWidgets.QMenu(self)
#         self.setMenu(menu)
#
#         for cmd, enabled, icon, text in [
#                 ('CLEAR', self._debugmode==0, pyzo.icons.bug_delete, clearallbps),
#                 ('PM', self._debugmode==0, pyzo.icons.bug_error,
#                     translate('debug', 'Postmortem: debug from last traceback')),
#                 ('STOP', self._debugmode>0, pyzo.icons.debug_quit,
#                     translate('debug', 'Stop debugging')),
# #                 ('NEXT', self._debugmode==2, pyzo.icons.debug_next,
# #                     translate('debug', 'Next: proceed until next line')),
# #                 ('STEP', self._debugmode==2, pyzo.icons.debug_step,
# #                     translate('debug', 'Step: proceed one step')),
# #                 ('RETURN', self._debugmode==2, pyzo.icons.debug_return,
# #                     translate('debug', 'Return: proceed until returns')),
# #                 ('CONTINUE', self._debugmode==2, pyzo.icons.debug_continue,
# #                     translate('debug', 'Continue: proceed to next breakpoint')),
#                 ]:
#             if cmd is None:
#                 menu.addSeparator()
#             else:
#                 if icon is not None:
#                     a = menu.addAction(icon, text)
#                 else:
#                     a = menu.addAction(text)
#                 if hasattr(text, 'tt'):
#                     a.setToolTip(text.tt)
#                 a.cmd = cmd
#                 a.setEnabled(enabled)
#
#
#     def onPressed(self, show=True):
#         self.buildMenu()
#         self.showMenu()
#
#
#     def onTriggered(self, action):
#         if action.cmd == 'PM':
#             # Initiate postmortem debugging
#             shell = pyzo.shells.getCurrentShell()
#             if shell:
#                 shell.executeCommand('DB START\n')
#
#         elif action.cmd == 'CLEAR':
#             # Clear all breakpoints
#             for e in pyzo.editors:
#                 e.clearBreakPoints()
#
#         else:
#             command = action.cmd.upper()
#             shell = pyzo.shells.getCurrentShell()
#             if shell:
#                 shell.executeCommand('DB %s\n' % command)
#
#
#     def setTrace(self, info):
#         """ Determine whether we are in debug mode.
#         """
#         if info is None:
#             self._debugmode = 0
#         else:
#             self._debugmode = info['debugmode']


class DebugStack(QtWidgets.QToolButton):
    """A button that shows the stack trace."""

    def __init__(self, parent):
        QtWidgets.QToolButton.__init__(self, parent)

        # Set text and tooltip
        self._baseText = translate("debug", "Stack")
        self.setText("%s:" % self._baseText)
        self.setIcon(pyzo.icons.text_align_justify)
        self.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        self.setPopupMode(self.InstantPopup)

        # Bind to triggers
        self.triggered.connect(self.onTriggered)

    def onTriggered(self, action):

        # Get shell
        shell = pyzo.shells.getCurrentShell()
        if not shell:
            return

        # Change stack index
        if not action._isCurrent:
            shell.executeCommand("DB FRAME {}\n".format(action._index))
        # Open file and select line
        if True:
            line = action.text().split(": ", 1)[1]
            self.debugFocus(line)

    def setTrace(self, info):
        """Set the stack trace. This method is called from
        the shell that receives the trace via its status channel
        directly from the interpreter.
        If trace is None, removes the trace
        """

        # Get info
        if info:
            index, frames, debugmode = info["index"], info["frames"], info["debugmode"]
        else:
            index, frames = -1, []

        if (not frames) or (debugmode == 0):

            # Remove trace
            self.setMenu(None)
            self.setText("")  # (self._baseText)
            self.setEnabled(False)
            pyzo.editors.setDebugLineIndicators(None)

        else:
            # Get the current frame
            theAction = None

            # Create menu and add __main__
            menu = QtWidgets.QMenu(self)
            self.setMenu(menu)

            # Fill trace
            for i in range(len(frames)):
                thisIndex = i + 1
                # Set text for action
                text = '{}: File "{}", line {}, in {}'
                text = text.format(thisIndex, *frames[i])
                action = menu.addAction(text)
                action._index = thisIndex
                action._isCurrent = False
                if thisIndex == index:
                    action._isCurrent = True
                    theAction = action
                    self.debugFocus(text.split(": ", 1)[1])  # Load editor

            # Get debug indicators
            debugIndicators = []
            for i in range(len(frames)):
                thisIndex = i + 1
                filename, linenr, func = frames[i]
                debugIndicators.append((filename, linenr))
                if thisIndex == index:
                    break
            # Set debug indicators
            pyzo.editors.setDebugLineIndicators(*debugIndicators)

            # Highlight current item and set the button text
            if theAction:
                menu.setDefaultAction(theAction)
                # self.setText(theAction.text().ljust(20))
                i = theAction._index
                text = "{} ({}/{}):  ".format(self._baseText, i, len(frames))
                self.setText(text)

            self.setEnabled(True)

    def debugFocus(self, lineFromDebugState):
        """debugFocus(lineFromDebugState)
        Open the file and show the linenr of the given lineFromDebugState.
        """
        # Get filenr and item
        try:
            tmp = lineFromDebugState.split(", in ")[0].split(", line ")
            filename = tmp[0][len("File ") :].strip('"')
            linenr = int(tmp[1].strip())
        except Exception:
            return "Could not focus!"
        # Cannot open <console>
        if filename == "<console>":
            return "Stack frame is <console>."
        elif filename.startswith("<ipython-input-"):
            return "Stack frame is IPython input."
        elif filename.startswith("<"):
            return "Stack frame is special name"
        # Go there!
        result = pyzo.editors.loadFile(filename)
        if not result:
            return "Could not open file where the error occured."
        else:
            editor = result._editor
            # Goto line and select it
            editor.gotoLine(linenr)
            cursor = editor.textCursor()
            cursor.movePosition(cursor.StartOfBlock)
            cursor.movePosition(cursor.EndOfBlock, cursor.KeepAnchor)
            editor.setTextCursor(cursor)


class InterpreterHelper(QtWidgets.QWidget):
    """This sits in place of a shell to help the user download miniconda."""

    def __init__(self, parent):
        super().__init__(parent)

        self._label = QtWidgets.QLabel("hello world")
        self._label.setTextFormat(QtCore.Qt.RichText)
        self._label.setWordWrap(True)
        # self._label.setOpenExternalLinks(True)
        self._label.linkActivated.connect(self.handle_link)
        font = self._label.font()
        font.setPointSize(font.pointSize() + 2)
        self._label.setFont(font)

        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self._label, 1)

    def refresh(self):
        self._label.setText("Detecting interpreters ...")
        QtWidgets.qApp.sendPostedEvents()
        QtWidgets.qApp.processEvents()
        self.detect()

    def detect(self):

        python_link = '<a href="https://www.python.org/">Python</a>'
        conda_link = '<a href="https://miniconda.pyzo.org">Miniconda</a>'
        self._the_exe = None
        configs = pyzo.config.shellConfigs2

        # Hide now?
        if configs and configs[0].exe:
            self._label.setText("Happy coding!")
            QtCore.QTimer.singleShot(1200, self.hide_this)
            return

        # Try to find an interpreter
        from pyzo.util.interpreters import get_interpreters

        interpreters = list(reversed(get_interpreters("2.4")))
        conda_interpreters = [i for i in interpreters if i.is_conda]
        conda_interpreters.sort(key=lambda x: len(x.path.replace("pyzo", "pyzo" * 10)))

        # Always sleep for a bit, so show that we've refreshed
        time.sleep(0.05)

        if conda_interpreters and conda_interpreters[0].version > "3":
            self._the_exe = conda_interpreters[0].path
            text = """Pyzo detected a conda environment in:
                      <br />%s<br /><br />
                      You can <a href='usefound'>use&nbsp;this&nbsp;environment</a>
                      (recommended), or manually specify an interpreter
                      by setting the exe in the <a href='config'>shell&nbsp;config</a>.
                      <br /><br />Click one of the links above, or <a href='refresh'>refresh</a>.
                   """ % (
                self._the_exe,
            )
        elif interpreters and interpreters[0].version > "3":
            self._the_exe = interpreters[0].path
            text = """Pyzo detected a Python interpreter in:
                      <br />%s<br /><br />
                      You can <a href='usefound'>use&nbsp;this&nbsp;environment</a>
                      (recommended), or manually specify an interpreter
                      by setting the exe in the <a href='config'>shell&nbsp;config</a>.
                      <br /><br />Click one of the links above, or <a href='refresh'>refresh</a>.
                   """ % (
                self._the_exe,
            )
        elif interpreters:
            text = """Pyzo detected a Python interpreter,
                      but it is Python 2. We strongly recommend using Python 3 instead.
                      <br /><br />
                      If you installed %s or %s in a non-default location,
                      or if you want to manually specify an interpreter,
                      set the exe in the <a href='config'>shell&nbsp;config</a>.
                      <br /><br />Click one of the links above, or <a href='refresh'>refresh</a>.
                   """ % (
                python_link,
                conda_link,
            )
        else:
            text = """Pyzo did not detect any Python interpreters.
                      We recomment installing %s or %s
                      (and click <a href='refresh'>refresh</a> when done).
                      <br /><br />
                      If you installed Python or Miniconda in a non-default location,
                      or if you want to manually specify the interpreter,
                      set the exe in the <a href='config'>shell&nbsp;config</a>.
                   """ % (
                python_link,
                conda_link,
            )

        link_style = "font-weight: bold; color:#369; text-decoration:underline;"
        self._label.setText(text.replace("<a ", '<a style="%s" ' % link_style))

    def handle_link(self, url):
        if url == "refresh":
            self.refresh()
        elif url == "config":
            self.editShellConfig()
        elif url == "usefound":
            self.useFound()
        elif url.startswith(("http://", "https://")):
            webbrowser.open(url)
        else:
            raise ValueError("Unknown link in conda helper: %s" % url)

    def editShellConfig(self):
        from pyzo.core.shellInfoDialog import ShellInfoDialog

        d = ShellInfoDialog()
        d.exec_()
        self.refresh()
        self.restart_shell()

    def useFound(self):
        # Set newfound interpreter
        if self._the_exe:
            configs = pyzo.config.shellConfigs2
            if not configs:
                from pyzo.core.kernelbroker import KernelInfo

                pyzo.config.shellConfigs2.append(KernelInfo())
            configs[0].exe = self._the_exe
            self.restart_shell()
        self.refresh()

    def hide_this(self):
        shells = self.parent()
        shells.showInterpreterHelper(False)

    def restart_shell(self):
        shells = self.parent()
        shell = shells.getCurrentShell()
        if shell is not None:
            shell.closeShell()
        shells.addShell(pyzo.config.shellConfigs2[0])



# -----------------------------------------------------------------------------
#  Copyright (C) 2013 Min RK
#
#  Distributed under the terms of the 2-clause BSD License.
# -----------------------------------------------------------------------------

from contextlib import contextmanager

import ctypes
import ctypes.util

objc = ctypes.cdll.LoadLibrary(ctypes.util.find_library("objc"))

void_p = ctypes.c_void_p
ull = ctypes.c_uint64

objc.objc_getClass.restype = void_p
objc.sel_registerName.restype = void_p
objc.objc_msgSend.restype = void_p
objc.objc_msgSend.argtypes = [void_p, void_p]

msg = objc.objc_msgSend


def _utf8(s):
    """ensure utf8 bytes"""
    if not isinstance(s, bytes):
        s = s.encode("utf8")
    return s


def n(name):
    """create a selector name (for methods)"""
    return objc.sel_registerName(_utf8(name))


def C(classname):
    """get an ObjC Class by name"""
    return objc.objc_getClass(_utf8(classname))


# constants from Foundation

NSActivityIdleDisplaySleepDisabled = 1 << 40
NSActivityIdleSystemSleepDisabled = 1 << 20
NSActivitySuddenTerminationDisabled = 1 << 14
NSActivityAutomaticTerminationDisabled = 1 << 15
NSActivityUserInitiated = 0x00FFFFFF | NSActivityIdleSystemSleepDisabled
NSActivityUserInitiatedAllowingIdleSystemSleep = (
    NSActivityUserInitiated & ~NSActivityIdleSystemSleepDisabled
)
NSActivityBackground = 0x000000FF
NSActivityLatencyCritical = 0xFF00000000


def beginActivityWithOptions(options, reason=""):
    """Wrapper for:

    [ [ NSProcessInfo processInfo]
        beginActivityWithOptions: (uint64)options
                          reason: (str)reason
    ]
    """
    NSProcessInfo = C("NSProcessInfo")
    NSString = C("NSString")

    reason = msg(NSString, n("stringWithUTF8String:"), _utf8(reason))
    info = msg(NSProcessInfo, n("processInfo"))
    activity = msg(
        info, n("beginActivityWithOptions:reason:"), ull(options), void_p(reason)
    )
    return activity


def endActivity(activity):
    """end a process activity assertion"""
    NSProcessInfo = C("NSProcessInfo")
    info = msg(NSProcessInfo, n("processInfo"))
    msg(info, n("endActivity:"), void_p(activity))


_theactivity = None


def nope():
    """disable App Nap by setting NSActivityUserInitiatedAllowingIdleSystemSleep"""
    global _theactivity
    _theactivity = beginActivityWithOptions(
        NSActivityUserInitiatedAllowingIdleSystemSleep, "Because Reasons"
    )


def nap():
    """end the caffeinated state started by `nope`"""
    global _theactivity
    if _theactivity is not None:
        endActivity(_theactivity)
        _theactivity = None


def napping_allowed():
    """is napping allowed?"""
    return _theactivity is None


@contextmanager
def nope_scope(
    options=NSActivityUserInitiatedAllowingIdleSystemSleep, reason="Because Reasons"
):
    """context manager for beginActivityWithOptions.

    Within this context, App Nap will be disabled.
    """
    activity = beginActivityWithOptions(options, reason)
    try:
        yield
    finally:
        endActivity(activity)


__all__ = [
    "NSActivityIdleDisplaySleepDisabled",
    "NSActivityIdleSystemSleepDisabled",
    "NSActivitySuddenTerminationDisabled",
    "NSActivityAutomaticTerminationDisabled",
    "NSActivityUserInitiated",
    "NSActivityUserInitiatedAllowingIdleSystemSleep",
    "NSActivityBackground",
    "NSActivityLatencyCritical",
    "beginActivityWithOptions",
    "endActivity",
    "nope",
    "nap",
    "napping_allowed",
    "nope_scope",
]



# -*- coding: utf-8 -*-
# Copyright (C) 2013, the Pyzo development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.


""" Package tools of pyzo

A tool consists of a module which contains a class. The id of
a tool is its module name made lower case. The module should
contain a class corresponding to its id. We advise to follow the
common python style and start the class name with a capital
letter, case does not matter for the tool to work though.
For instance, the tool "pyzologger" is the class "PyzoLogger" found
in module "pyzoLogger"

The module may contain the following extra variables (which should
be placed within the first 50 lines of code):

tool_name - A readable name for the tool (may contain spaces,
will be shown in the tab)

tool_summary - A single line short summary of the tool. To be
displayed in the statusbar.
"""

# tools I'd like:
# - find in files
# - workspace
# - source tree
# - snipet manager
# - file browser
# - pythonpath editor, startupfile editor (or as part of pyzo?)

import os, sys, imp

import pyzo
from pyzo.qt import QtCore, QtGui, QtWidgets  # noqa
from pyzo.util import zon as ssdf
from pyzo import translate  # noqa (we have an eval down here)


class ToolDockWidget(QtWidgets.QDockWidget):
    """A dock widget that holds a tool.
    It sets all settings, initializes the tool widget, and notifies the
    tool manager on closing.
    """

    def __init__(self, parent, toolManager):
        QtWidgets.QDockWidget.__init__(self, parent)

        # Store stuff
        self._toolManager = toolManager

        # Allow docking anywhere, othwerise restoring state wont work properly

        # Set other settings
        self.setFeatures(
            QtWidgets.QDockWidget.DockWidgetMovable
            | QtWidgets.QDockWidget.DockWidgetClosable
            | QtWidgets.QDockWidget.DockWidgetFloatable
            # QtWidgets.QDockWidget.DockWidgetVerticalTitleBar
        )

    def setTool(self, toolId, toolName, toolClass):
        """Set the tool information. Call this right after
        initialization."""

        # Store id and set object name to enable saving/restoring state
        self._toolId = toolId
        self.setObjectName(toolId)

        # Set name
        self.setWindowTitle(toolName)

        # Create tool widget
        self.reload(toolClass)

    def closeEvent(self, event):
        if self._toolManager:
            self._toolManager.onToolClose(self._toolId)
            self._toolManager = None
        # Close and delete widget
        old = self.widget()
        if old:
            old.close()
            old.deleteLater()
        # Close and delete dock widget
        self.close()
        self.deleteLater()
        # We handled the event
        event.accept()

    def reload(self, toolClass):
        """Reload the widget with a new widget class."""
        old = self.widget()
        new = toolClass(pyzo.main)
        self.setWidget(new)
        if old:
            old.close()
            old.deleteLater()


class ToolDescription:
    """Provides a description of a tool and has a reference to
    the tool dock instance if it is loaded.
    """

    def __init__(self, modulePath, name="", description=""):
        # Set names
        self.modulePath = modulePath
        self.moduleName = os.path.splitext(os.path.basename(modulePath))[0]
        self.id = self.moduleName.lower()
        if name:
            self.name = name
        else:
            self.name = self.id

        # Set description
        self.description = description
        # Init instance to None, will be set when loaded
        self.instance = None

    def menuLauncher(self, value):
        """Function that is called by the menu when this tool is selected."""
        if value is None:
            return bool(self.instance)
            # return self.id in pyzo.toolManager._activeTools
        elif value:
            pyzo.toolManager.loadTool(self.id)
        else:
            self.widget = None
            pyzo.toolManager.closeTool(self.id)


class ToolManager(QtCore.QObject):
    """Manages the tools."""

    # This signal indicates a change in the loaded tools
    toolInstanceChange = QtCore.Signal()

    def __init__(self, parent=None):
        QtCore.QObject.__init__(self, parent)

        # list of ToolDescription instances
        self._toolInfo = None
        self._activeTools = {}

    def loadToolInfo(self):
        """(re)load the tool information."""
        # Get paths to load files from
        toolDir1 = os.path.join(pyzo.pyzoDir, "tools")
        toolDir2 = os.path.join(pyzo.appDataDir, "tools")

        # Create list of tool files
        toolfiles = []
        for toolDir in [toolDir1, toolDir2]:
            tmp = [os.path.join(toolDir, f) for f in os.listdir(toolDir)]
            toolfiles.extend(tmp)

        # Note: we do not use the code below anymore, since even the frozen
        # app makes use of the .py files.
        #         # Get list of files, also when we're in a zip file.
        #         i = tooldir.find('.zip')
        #         if i>0:
        #             # Get list of files from zipfile
        #             tooldir = tooldir[:i+4]
        #             import zipfile
        #             z = zipfile.ZipFile(tooldir)
        #             toolfiles = [os.path.split(i)[1] for i in z.namelist()
        #                         if i.startswith('visvis') and i.count('functions')]
        #         else:
        #             # Get list of files from file system
        #             toolfiles = os.listdir(tooldir)

        # Iterate over tool modules
        newlist = []
        for file in toolfiles:
            modulePath = file

            # Check
            if os.path.isdir(file):
                file = os.path.join(file, "__init__.py")  # A package perhaps
                if not os.path.isfile(file):
                    continue
            elif file.endswith("__.py") or not file.endswith(".py"):
                continue
            elif file.endswith("pyzoFileBrowser.py"):
                # Skip old file browser (the file can be there from a previous install)
                continue

            #
            toolName = ""
            toolSummary = ""
            # read file to find name or summary
            linecount = 0
            for line in open(file, encoding="utf-8"):
                linecount += 1
                if linecount > 50:
                    break
                if line.startswith("tool_name"):
                    i = line.find("=")
                    if i < 0:
                        continue
                    line = line.rstrip("\n").rstrip("\r")
                    line = line[i + 1 :].strip(" ")
                    toolName = eval(line)  # applies translation
                elif line.startswith("tool_summary"):
                    i = line.find("=")
                    if i < 0:
                        continue
                    line = line.rstrip("\n").rstrip("\r")
                    line = line[i + 1 :].strip(" ")
                    toolSummary = line.strip("'").strip('"')
                else:
                    pass

            # Add stuff
            tmp = ToolDescription(modulePath, toolName, toolSummary)
            newlist.append(tmp)

        # Store and return
        self._toolInfo = sorted(newlist, key=lambda x: x.id)
        self.updateToolInstances()
        return self._toolInfo

    def updateToolInstances(self):
        """Make tool instances up to date, so that it can be seen what
        tools are now active."""
        for toolDes in self.getToolInfo():
            if toolDes.id in self._activeTools:
                toolDes.instance = self._activeTools[toolDes.id]
            else:
                toolDes.instance = None

        # Emit update signal
        self.toolInstanceChange.emit()

    def getToolInfo(self):
        """Like loadToolInfo(), but use buffered instance if available."""
        if self._toolInfo is None:
            self.loadToolInfo()
        return self._toolInfo

    def getToolClass(self, toolId):
        """Get the class of the tool.
        It will import (and reload) the module and get the class.
        Some checks are performed, like whether the class inherits
        from QWidget.
        Returns the class or None if failed...
        """

        # Make sure we have the info
        if self._toolInfo is None:
            self.loadToolInfo()

        # Get module name and path
        for toolDes in self._toolInfo:
            if toolDes.id == toolId:
                moduleName = toolDes.moduleName
                modulePath = toolDes.modulePath
                break
        else:
            print("WARNING: could not find module for tool", repr(toolId))
            return None

        # Remove from sys.modules, to force the module to reload
        for key in [key for key in sys.modules]:
            if key and key.startswith("pyzo.tools." + moduleName):
                del sys.modules[key]

        # Load module
        try:
            m_file, m_fname, m_des = imp.find_module(
                moduleName, [os.path.dirname(modulePath)]
            )
            mod = imp.load_module("pyzo.tools." + moduleName, m_file, m_fname, m_des)
        except Exception as why:
            print("Invalid tool " + toolId + ":", why)
            return None

        # Is the expected class present?
        className = ""
        for member in dir(mod):
            if member.lower() == toolId:
                className = member
                break
        else:
            print("Invalid tool, Classname must match module name '%s'!" % toolId)
            return None

        # Does it inherit from QWidget?
        plug = mod.__dict__[className]
        if not (isinstance(plug, type) and issubclass(plug, QtWidgets.QWidget)):
            print("Invalid tool, tool class must inherit from QWidget!")
            return None

        # Succes!
        return plug

    def loadTool(self, toolId, splitWith=None):
        """Load a tool by creating a dock widget containing the tool widget."""

        # A tool id should always be lower case
        toolId = toolId.lower()

        # Close old one
        if toolId in self._activeTools:
            old = self._activeTools[toolId].widget()
            self._activeTools[toolId].setWidget(QtWidgets.QWidget(pyzo.main))
            if old:
                old.close()
                old.deleteLater()

        # Get tool class (returns None on failure)
        toolClass = self.getToolClass(toolId)
        if toolClass is None:
            return

        # Already loaded? reload!
        if toolId in self._activeTools:
            self._activeTools[toolId].reload(toolClass)
            return

        # Obtain name from buffered list of names
        for toolDes in self._toolInfo:
            if toolDes.id == toolId:
                name = toolDes.name
                break
        else:
            name = toolId

        # Make sure there is a config entry for this tool
        if not hasattr(pyzo.config.tools, toolId):
            pyzo.config.tools[toolId] = ssdf.new()

        # Create dock widget and add in the main window
        dock = ToolDockWidget(pyzo.main, self)
        dock.setTool(toolId, name, toolClass)

        if splitWith and splitWith in self._activeTools:
            otherDock = self._activeTools[splitWith]
            pyzo.main.splitDockWidget(otherDock, dock, QtCore.Qt.Horizontal)
        else:
            pyzo.main.addDockWidget(QtCore.Qt.RightDockWidgetArea, dock)

        # Add to list
        self._activeTools[toolId] = dock
        self.updateToolInstances()

    def reloadTools(self):
        """Reload all tools."""
        for id in self.getLoadedTools():
            self.loadTool(id)

    def closeTool(self, toolId):
        """Close the tool with specified id."""
        if toolId in self._activeTools:
            dock = self._activeTools[toolId]
            dock.close()

    def getTool(self, toolId):
        """Get the tool widget instance, or None
        if not available."""
        if toolId in self._activeTools:
            return self._activeTools[toolId].widget()
        else:
            return None

    def onToolClose(self, toolId):
        # Remove from dict
        self._activeTools.pop(toolId, None)
        # Set instance to None
        self.updateToolInstances()

    def getLoadedTools(self):
        """Get a list with id's of loaded tools."""
        tmp = []
        for toolDes in self.getToolInfo():
            if toolDes.id in self._activeTools:
                tmp.append(toolDes.id)
        return tmp



# -*- coding: utf-8 -*-
# Copyright (C) 2013, Almar Klein

# From pyzo/pyzokernel

import sys
import time
import subprocess


def subprocess_with_callback(callback, cmd, **kwargs):
    """Execute command in subprocess, stdout is passed to the
    callback function. Returns the returncode of the process.
    If callback is None, simply prints any stdout.
    """

    # Set callback to void if None
    if callback is None:
        callback = lambda x: None

    # Execute command
    try:
        p = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, **kwargs
        )
    except OSError:
        type, err, tb = sys.exc_info()
        del tb
        callback(str(err) + "\n")
        return -1

    pending = []
    while p.poll() is None:
        time.sleep(0.001)
        # Read text and process
        c = p.stdout.read(1).decode("utf-8", "ignore")
        pending.append(c)
        if c in ".\n":
            callback("".join(pending))
            pending = []

    # Process remaining text
    pending.append(p.stdout.read().decode("utf-8"))
    callback("".join(pending))

    # Done
    return p.returncode


def print_(p):
    sys.stdout.write(p)
    sys.stdout.flush()


def pip_command_exe(exe, *args):
    """Do a pip command in the interpreter with the given exe."""

    # Get pip command
    cmd = [exe, "-m", "pip"] + list(args)

    # Execute it
    subprocess_with_callback(print_, cmd)


def pip_command(*args):
    """Do a pip command, e.g. "install networkx".
    Installs in the current interpreter.
    """
    pip_command_exe(sys.executable, *args)


if __name__ == "__main__":
    pip_command("install", "networkx")
